"""
Module containing reduction raster blocks.
"""
import numpy as np
from dask_geomodeling.utils import get_index, parse_percentile_statistic

from functools import partial

STATISTICS = {
    "first": None,
    "last": None,
    "count": None,
    "sum": np.nansum,
    "mean": np.nanmean,
    "min": np.nanmin,
    "max": np.nanmax,
    "argmin": np.nanargmin,
    "argmax": np.nanargmax,
    "std": np.nanstd,
    "var": np.nanvar,
    "median": np.nanmedian,
    # "p<number>" uses np.nanpercentile
}


def check_statistic(statistic):
    if statistic not in STATISTICS:
        percentile = parse_percentile_statistic(statistic)
        if percentile is None:
            raise ValueError('Unknown statistic "{}"'.format(statistic))


def reduce_rasters(stack, statistic, no_data_value=None, dtype=None):
    """Apply a statistic (e.g. "mean") to a stack of rasters, skipping
    'no data' values.

    In this context, reduce means that the dimensionality of the input data
    is reduced by one.

    Args:
      stack (list): a list of dicts containing "values" (ndarray)
        and "no_data_value". If the list has zero length or if the ndarrays
        do not have the same shape, a ValueError is raised.
      statistic (str): the applied statistic (no data is ignored). One of:
        {"last", "first", "count", "sum", "mean", "min",
        "max", "argmin", "argmax", "product", "std", "var", "p<number>"}
      no_data_value (number, optional): the 'no data' value in the output
        array. Defaults to the no data value of the first element in the stack.
      dtype (str or dtype): the datatype of the output array. Defaults to the
        dtype of the first element in the stack. If the input
        data cannot be cast to this dtype, a ValueError is raised.

    Returns:
      dict with "values" and "no_data_value"
    """
    if statistic not in STATISTICS:
        percentile = parse_percentile_statistic(statistic)
        if percentile is None:
            raise KeyError('Unknown statistic "{}"'.format(statistic))
        else:
            statistic = "percentile"

    if len(stack) == 0:
        raise ValueError("Cannot reduce a zero-length stack")

    # get the output array properties (dtype, no_data_value, shape)
    if dtype is None:
        dtype = stack[0]["values"].dtype
    if no_data_value is None:
        no_data_value = stack[0]["no_data_value"]
    shape = stack[0]["values"].shape

    # sum, count and nans output do not contain no data: fill zeroes right away
    if statistic in {"sum", "count", "nans"}:
        fill_value = 0
    else:
        fill_value = no_data_value

    # create the output array
    out = np.full(shape, fill_value, dtype)

    if statistic == "last":
        # populate 'out' with the last value that is not 'no data'
        for data in stack:
            index = get_index(data["values"], data["no_data_value"])
            out[index] = data["values"][index]
    elif statistic == "first":
        # populate 'out' with the first value that is not 'no data'
        for data in stack[::-1]:
            index = get_index(data["values"], data["no_data_value"])
            out[index] = data["values"][index]
    elif statistic == "count":
        # count the number of values that are not 'no data'
        for data in stack:
            out += get_index(data["values"], data["no_data_value"])
    else:
        if statistic == "percentile":
            func = partial(np.nanpercentile, q=percentile)
        else:
            func = STATISTICS[statistic]
        # transform 'no data' into 'nan' to be able to use numpy functions
        # NB: the dtype is at least float16 to accomodate NaN
        stack_array = np.full(
            (len(stack),) + shape, np.nan, np.result_type(dtype, np.float16)
        )
        for i, data in enumerate(stack):
            index = get_index(data["values"], data["no_data_value"])
            stack_array[i, index] = data["values"][index]

        # protect against all-NaN slice warnings and errors
        not_all_nan = ~np.all(np.isnan(stack_array), axis=0)

        # perform the math
        out[not_all_nan] = func(stack_array[:, not_all_nan], axis=0)

    return {"values": out, "no_data_value": no_data_value}



"""
Module containing raster blocks that aggregate rasters.
"""
from math import ceil, floor, log, sqrt
from collections import defaultdict
from functools import partial
import warnings

from scipy import ndimage
import numpy as np
import geopandas as gpd

from dask import config
from dask_geomodeling import measurements
from dask_geomodeling import utils
from dask_geomodeling.raster import RasterBlock

from .base import GeometryBlock

__all__ = ["AggregateRaster", "AggregateRasterAboveThreshold"]


class Bucket:
    """
    Track objects in an imaginary grid that may span up to 4 cells.
    """

    def __init__(self):
        self.cells = set()
        self.indices = []

    def __contains__(self, cells):
        """
        Return wether any of the cells defined by indices is already occupied.
        """
        return bool(self.cells & cells)

    def add(self, index, cells):
        """
        Update the set of occupied cells with cells and append index to the
        list of indices.

        Note that this does not fail if cells intersects with already occupied
        cells.
        """
        self.indices.append(index)
        self.cells.update(cells)


def calculate_level_and_cells(bbox):
    """
    Return a tuple (level, cells).

    :param bboxes: list of (xmin, ymin, xmax, ymax) tuples

    The returned cells is a set of indices which represent occupied cells (at
    most 4) in an imaginary sparse grid that has a cellsize defined by the
    integer level.  Level 0 corresponds to the unit cell. Each doubling of the
    cellsize level increase corresponds to a doubling of the cellsize of the
    previous level.
    """
    x1, y1, x2, y2 = bbox
    level = -ceil(log(max(x2 - x1, y2 - y1), 2))

    width = 0.5 ** level
    height = 0.5 ** level

    j1 = floor(x1 / width)
    j2 = floor(x2 / width)
    i1 = floor(y1 / height)
    i2 = floor(y2 / height)

    return level, {(i1, j1), (i1, j2), (i2, j1), (i2, j2)}


def bucketize(bboxes):
    """
    Return list of lists with indices into bboxes.

    :param bboxes: list of (xmin, ymin, xmax, ymax) tuples

    Each sublist in the returned list points to a subset of disjoint bboxes.
    Instead of aiming for the smallest amount of subsets possible, this
    approach focuses on speed by avoiding costly intersection operations on all
    bboxes in a bucket.
    """
    bucket_dict = defaultdict(list)

    for index, bbox in enumerate(bboxes):
        level, cells = calculate_level_and_cells(bbox)

        # select bucket by feature size
        bucket_list = bucket_dict[level]

        for bucket in bucket_list:
            # see if it fits in this bucket
            if cells in bucket:
                continue
            # a suitable bucket has been found, break out of the for loop
            break
        else:
            # no break, no suitable bucket found, assign and append a new one
            bucket = Bucket()
            bucket_list.append(bucket)

        # add the item to the bucket
        bucket.add(index=index, cells=cells)

    return [
        bucket.indices for bucket_list in bucket_dict.values() for bucket in bucket_list
    ]


class AggregateRaster(GeometryBlock):
    """
    Compute statistics of a raster for each geometry in a geometry source.

    A statistic is computed in a specific projection and with a specified cell
    size. If ``projection`` or ``pixel_size`` are not given, these default to
    the native projection of the provided raster source.

    Should the combination of the requested pixel_size and the extent of the
    source geometry cause the required raster size to exceed ``max_pixels``,
    the ``pixel_size`` can be adjusted automatically if ``auto_pixel_size`` is
    set to ``True``, else (the default) a RuntimeError is raised.

    Please note that for any field operation on the result of this block
    a GetSeriesBlock should be used to retrieve data from the added column. The
    name of the added column is determined by the ``column_name`` parameter.

    Args:
      source (GeometryBlock): The geometry source for which the statistics are
        determined.
      raster (RasterBlock): The raster source that is sampled.
      statistic (str): The type of statistical analysis that should be
        performed. The options are: ``{"sum", "count", "min", "max", "mean",
        "median", "p<percentile>"}``. Percentiles are provided for example as
        follows: ``"p50"``. Default ``"sum"``.
      projection (str, optional): Projection to perform the aggregation in, for
        example ``"EPSG:28992"``. Defaults to the native projection of the
        supplied raster.
      pixel_size (float, optional): The raster cell size used in the
        aggregation. Defaults to the cell size of the supplied raster.
      max_pixels (int, optional): The maximum number of pixels (cells) in the
        aggregation. Defaults to the ``geomodeling.raster-limit`` setting.
      column_name (str, optional): The name of the column where the result
        should be placed. Defaults to ``"agg"``.
      auto_pixel_size (boolean): Determines whether the pixel size is adjusted
        automatically when ``"max_pixels"`` is exceeded. Default False.

    Returns: 
      GeometryBlock with aggregation results in an added column

    The global raster-limit setting can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.raster-limit": 10 ** 9})
    """

    # extensive (opposite: intensive) means: additive, proportional to size
    STATISTICS = {
        "sum": {"func": ndimage.sum, "extensive": True},
        "count": {"func": ndimage.sum, "extensive": True},
        "min": {"func": ndimage.minimum, "extensive": False},
        "max": {"func": ndimage.maximum, "extensive": False},
        "mean": {"func": ndimage.mean, "extensive": False},
        "median": {"func": ndimage.median, "extensive": False},
        "percentile": {"func": measurements.percentile, "extensive": False},
    }

    def __init__(
        self,
        source,
        raster,
        statistic="sum",
        projection=None,
        pixel_size=None,
        max_pixels=None,
        column_name="agg",
        auto_pixel_size=False,
        *args
    ):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(raster, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(raster)))
        if not isinstance(statistic, str):
            raise TypeError("'{}' object is not allowed".format(type(statistic)))
        statistic = statistic.lower()
        percentile = utils.parse_percentile_statistic(statistic)
        if percentile:
            statistic = "p{0}".format(percentile)
        elif statistic not in self.STATISTICS or statistic == "percentile":
            raise ValueError("Unknown statistic '{}'".format(statistic))

        if projection is None:
            projection = raster.projection
        if not isinstance(projection, str):
            raise TypeError("'{}' object is not allowed".format(type(projection)))
        if pixel_size is None:
            # get the pixel_size from the raster geo_transform
            geo_transform = raster.geo_transform
            if geo_transform is None:
                raise ValueError(
                    "Cannot get the pixel_size from the source "
                    "raster. Please provide a pixel_size."
                )
            pixel_size = min(abs(float(geo_transform[1])), abs(float(geo_transform[5])))
        else:
            pixel_size = abs(float(pixel_size))
        if pixel_size == 0.0:
            raise ValueError("Pixel size cannot be 0")
        if max_pixels is not None:
            max_pixels = int(max_pixels)
        if not isinstance(auto_pixel_size, bool):
            raise TypeError("'{}' object is not allowed".format(type(auto_pixel_size)))

        super(AggregateRaster, self).__init__(
            source,
            raster,
            statistic,
            projection,
            pixel_size,
            max_pixels,
            column_name,
            auto_pixel_size,
            *args
        )

    @property
    def source(self):
        return self.args[0]

    @property
    def raster(self):
        return self.args[1]

    @property
    def statistic(self):
        return self.args[2]

    @property
    def projection(self):
        return self.args[3]

    @property
    def pixel_size(self):
        return self.args[4]

    @property
    def max_pixels(self):
        return self.args[5]

    @property
    def column_name(self):
        return self.args[6]

    @property
    def auto_pixel_size(self):
        return self.args[7]

    @property
    def columns(self):
        return self.source.columns | {self.column_name}

    def get_sources_and_requests(self, **request):
        if request.get("mode") == "extent":
            return [(self.source, request), (None, None), ({"mode": "extent"}, None)]

        req_srs = request["projection"]
        agg_srs = self.projection

        # acquire the extent of the geometry data
        extent_request = {**request, "mode": "extent"}
        extent = self.source.get_data(**extent_request)["extent"]

        if extent is None:
            # make sources_and_request so that we get an empty result
            return [
                (None, None),
                (None, None),
                ({"empty": True, "projection": req_srs}, None),
            ]

        # transform the extent into the projection in which we aggregate
        x1, y1, x2, y2 = utils.transform_extent(extent, req_srs, agg_srs)

        # estimate the amount of required pixels
        required_pixels = int(((x2 - x1) * (y2 - y1)) / (self.pixel_size ** 2))

        # in case this request is too large, we adapt pixel size
        max_pixels = self.max_pixels
        if max_pixels is None:
            max_pixels = config.get("geomodeling.raster-limit")
        pixel_size = self.pixel_size

        if required_pixels > max_pixels and self.auto_pixel_size:
            # adapt with integer multiples of pixel_size
            pixel_size *= ceil(sqrt(required_pixels / max_pixels))
        elif required_pixels > max_pixels:
            raise RuntimeError(
                "The required raster size for the aggregation exceeded "
                "the maximum ({} > {})".format(required_pixels, max_pixels)
            )

        # snap the extent to (0, 0) to prevent subpixel shifts
        x1 = floor(x1 / pixel_size) * pixel_size
        y1 = floor(y1 / pixel_size) * pixel_size
        x2 = ceil(x2 / pixel_size) * pixel_size
        y2 = ceil(y2 / pixel_size) * pixel_size

        # compute the width and height
        width = max(int((x2 - x1) / pixel_size), 1)
        height = max(int((y2 - y1) / pixel_size), 1)

        raster_request = {
            "mode": "vals",
            "projection": agg_srs,
            "start": request.get("start"),
            "stop": request.get("stop"),
            "aggregation": None,  # TODO
            "bbox": (x1, y1, x2, y2),
            "width": width,
            "height": height,
        }

        process_kwargs = {
            "mode": request.get("mode", "intersects"),
            "pixel_size": self.pixel_size,
            "agg_srs": agg_srs,
            "req_srs": req_srs,
            "actual_pixel_size": pixel_size,
            "statistic": self.statistic,
            "result_column": self.column_name,
            "agg_bbox": (x1, y1, x2, y2),
        }

        return [
            (self.source, request),
            (self.raster, raster_request),
            (process_kwargs, None),
        ]

    @staticmethod
    def process(geom_data, raster_data, process_kwargs):
        if process_kwargs.get("empty"):
            return {
                "features": gpd.GeoDataFrame([]),
                "projection": process_kwargs["projection"],
            }
        elif process_kwargs["mode"] == "extent":
            return geom_data

        features = geom_data["features"]
        if len(features) == 0:
            return geom_data

        result = features.copy()

        # transform the features into the aggregation projection
        req_srs = process_kwargs["req_srs"]
        agg_srs = process_kwargs["agg_srs"]

        agg_geometries = utils.geoseries_transform(
            features["geometry"], req_srs, agg_srs,
        )

        statistic = process_kwargs["statistic"]
        percentile = utils.parse_percentile_statistic(statistic)
        if percentile:
            statistic = "percentile"
            agg_func = partial(
                AggregateRaster.STATISTICS[statistic]["func"], qval=percentile
            )
        else:
            agg_func = AggregateRaster.STATISTICS[statistic]["func"]

        extensive = AggregateRaster.STATISTICS[statistic]["extensive"]
        result_column = process_kwargs["result_column"]

        # this is only there for the AggregateRasterAboveThreshold
        threshold_name = process_kwargs.get("threshold_name")
        if threshold_name:
            # get the threshold, appending NaN for unlabeled pixels
            threshold_values = np.empty((len(features) + 1, ), dtype="f4")
            threshold_values[:-1] = features[threshold_name].values
            threshold_values[-1] = np.nan
        else:
            threshold_values = None

        # investigate the raster data
        if raster_data is None:
            values = no_data_value = None
        else:
            values = raster_data["values"]
            no_data_value = raster_data["no_data_value"]
        if values is None or np.all(values == no_data_value):  # skip the rest
            result[result_column] = 0 if extensive else np.nan
            return {"features": result, "projection": req_srs}
        depth, height, width = values.shape

        pixel_size = process_kwargs["pixel_size"]
        actual_pixel_size = process_kwargs["actual_pixel_size"]

        # process in groups of disjoint subsets of the features
        agg = np.full((depth, len(features)), np.nan, dtype="f4")
        for select in bucketize(features.bounds.values):
            rasterize_result = utils.rasterize_geoseries(
                agg_geometries.iloc[select],
                process_kwargs["agg_bbox"],
                agg_srs,
                height,
                width,
                values=np.asarray(select, dtype=np.int32),  # GDAL needs int32
            )
            labels = rasterize_result["values"][0]

            # if there is a threshold, generate a raster with thresholds
            if threshold_name:
                # mode="clip" ensures that unlabeled cells use the appended NaN
                thresholds = np.take(threshold_values, labels, mode="clip")
            else:
                thresholds = None

            for frame_no, frame in enumerate(values):
                # limit statistics to active pixels
                active = frame != no_data_value
                # if there is a threshold, mask the frame
                if threshold_name:
                    valid = ~np.isnan(thresholds)  # to suppress warnings
                    active[~valid] = False  # no threshold -> no aggregation
                    active[valid] &= frame[valid] >= thresholds[valid]

                # if there is no single active value: do not aggregate
                if not active.any():
                    continue

                # select features that actually have data
                # (min, max, median, and percentile cannot handle it otherwise)
                active_labels = labels[active]
                select_and_active = list(
                    set(np.unique(active_labels)) & set(select)
                )

                if not select_and_active:
                    continue

                agg[frame_no][select_and_active] = agg_func(
                    1 if statistic == "count" else frame[active],
                    labels=active_labels,
                    index=select_and_active,
                )

        if extensive:  # sum and count
            agg[~np.isfinite(agg)] = 0
            # extensive aggregations have to be scaled
            if actual_pixel_size != pixel_size:
                agg *= (actual_pixel_size / pixel_size) ** 2
        else:
            agg[~np.isfinite(agg)] = np.nan  # replaces inf by nan

        if depth == 1:
            result[result_column] = agg[0]
        else:
            # store an array in a dataframe cell: set each cell with [np.array]
            result[result_column] = [[x] for x in agg.T]

        return {"features": result, "projection": req_srs}


class AggregateRasterAboveThreshold(AggregateRaster):
    """
    Compute statistics of a per-feature masked raster for each geometry in a
    geometry source.

    Per feature, a threshold can be supplied to mask the raster with. Only
    values that exceed the threshold of a specific feature are included for
    the statistical value of that feature.

    See :class:``dask_geomodeling.geometry.aggregate.AggregateRaster`` for
    further information.

    Args:
      *args: See :class:``dask_geomodeling.geometry.aggregate.AggregateRaster``
      threshold_name (str): The column that holds the thresholds.

    Returns:
      GeometryBlock with aggregation results in an added column
    """
    def __init__(
        self,
        source,
        raster,
        statistic="sum",
        projection=None,
        pixel_size=None,
        max_pixels=None,
        column_name="agg",
        auto_pixel_size=False,
        threshold_name=None,
    ):
        if not isinstance(threshold_name, str):
            raise TypeError("'{}' object is not allowed".format(type(threshold_name)))
        if threshold_name not in source.columns:
            raise KeyError("Column '{}' is not available".format(threshold_name))
        super().__init__(
            source,
            raster,
            statistic,
            projection,
            pixel_size,
            max_pixels,
            column_name,
            auto_pixel_size,
            threshold_name,
        )

    @property
    def threshold_name(self):
        return self.args[8]

    def get_sources_and_requests(self, **request):
        src_and_req = super().get_sources_and_requests(**request)
        process_kwargs = src_and_req[2][0]
        process_kwargs["threshold_name"] = self.threshold_name
        return src_and_req



from datetime import datetime
from io import BytesIO
from urllib.parse import urljoin

import numpy as np
import traitlets
from ipyleaflet import Map, WMSLayer
from matplotlib import cm
from matplotlib.colors import Normalize
from notebook import notebookapp
from notebook.base.handlers import IPythonHandler
from notebook.utils import url_path_join
from PIL import Image

from dask_geomodeling.core import Block


class GeomodelingWMSHandler(IPythonHandler):
    """This Tornado request handler adds a WMS functionality for displaying
    dask-geomodeling results in a notebook

    See:
    https://jupyter-notebook.readthedocs.io/en/stable/extending/handlers.html
    """

    def get(self):
        block = Block.from_json(self.get_query_argument("layers"))
        style = self.get_query_argument("styles")
        vmin = float(self.get_query_argument("vmin"))
        vmax = float(self.get_query_argument("vmax"))
        format = self.get_query_argument("format")
        if format.lower() != "image/png":
            self.set_status(400)
            self.finish("Only image/png is supported")
            return
        srs = self.get_query_argument("srs")
        height = int(self.get_query_argument("height"))
        max_cell_size = float(self.get_query_argument("maxcellsize"))
        time_isoformat = self.get_query_argument("time")
        if time_isoformat:
            time = datetime.strptime(time_isoformat, "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            time = None
        width = int(self.get_query_argument("width"))
        bbox = [float(x) for x in self.get_query_argument("bbox").split(",")]

        # overload protection
        cell_size_x = (bbox[2] - bbox[0]) / width
        cell_size_y = (bbox[3] - bbox[1]) / height
        if cell_size_x > max_cell_size or cell_size_y > max_cell_size:
            self.set_status(400)
            self.finish("Too large area requested")
            return

        # get cmap
        data = block.get_data(
            mode="vals",
            bbox=bbox,
            height=height,
            width=width,
            projection=srs,
            start=time,
        )
        masked = np.ma.masked_equal(data["values"][0], data["no_data_value"])
        stream = BytesIO()

        normalized = Normalize(vmin=vmin, vmax=vmax, clip=True)(masked)
        img = cm.get_cmap(style)(normalized)
        img[normalized.mask, 3] = 0.0
        img_uint8 = (img * 255).astype(np.uint8)
        Image.fromarray(img_uint8).save(stream, format="png")
        raw = stream.getvalue()

        self.set_header("Content-Length", len(raw))
        self.set_header("Content-Type", "image/png")
        self.set_header("Pragma", "no-cache")
        self.set_header(
            "Cache-Control",
            "no-store, "
            "no-cache=Set-Cookie, "
            "proxy-revalidate, "
            "max-age=0, "
            "post-check=0, pre-check=0",
        )
        self.set_header("Expires", "Wed, 2 Dec 1837 21:00:12 GMT")
        self.write(raw)
        self.finish()


class GeomodelingLayer(WMSLayer):
    """Visualize a dask_geomodeling.RasterBlock on a ipyleaflet Map.

    :param block: a dask_geomodeling.RasterBlock instance to visualize
    :param url: The url of the jupyter server (e.g. https://localhost:8888)
    :param style: a valid matplotlib colormap
    :param vmin: the minimum value (for the colormap)
    :param vmax: the maximum value (for the colormap)

    Notes
    -----
    To use this ipyleaflet extension, you have to include this plugin into your
    Jupyter Notebook server by calling::

    $ jupyter notebook --NotebookApp.nbserver_extensions="{'dask_geomodeling.ipyleaflet_plugin':True}"

    Or, by adding this setting to the config file in ~/.jupyter, which can be
    generated by calling::

    $ jupyter notebook --generate-config

    This plugin extends the Jupyter notebook server with a server that responds
    with PNG images for WMS requests generated by ipyleaflet.
    """

    format = traitlets.Unicode("image/png").tag(sync=True, o=True)
    maxcellsize = traitlets.Float(10.0).tag(sync=True, o=True)
    time = traitlets.Unicode("").tag(sync=True, o=True)
    vmin = traitlets.Float(0.0).tag(sync=True, o=True)
    vmax = traitlets.Float(1.0).tag(sync=True, o=True)

    def __init__(self, block, url=None, **kwargs):
        if url is None:
            # just get the first URL one
            url = next(notebookapp.list_running_servers())["url"]
        self.layers = block.to_json()
        super().__init__(url=urljoin(url, "wms"), **kwargs)


def load_jupyter_server_extension(nb_server_app):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """
    web_app = nb_server_app.web_app
    host_pattern = ".*$"
    route_pattern = url_path_join(web_app.settings["base_url"], "/wms")
    web_app.add_handlers(host_pattern, [(route_pattern, GeomodelingWMSHandler)])



"""
Module containing elementwise raster blocks.
"""
from functools import wraps

import numpy as np

from dask_geomodeling.utils import get_dtype_max, get_index, GeoTransform

from .base import RasterBlock, BaseSingle

__all__ = [
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Power",
    "FillNoData",
    "Equal",
    "NotEqual",
    "Greater",
    "GreaterEqual",
    "Less",
    "LessEqual",
    "Invert",
    "And",
    "Or",
    "Xor",
    "IsData",
    "IsNoData",
]


class BaseElementwise(RasterBlock):
    """ Base block for elementwise operations on rasters.

    This block only accepts stores that have aligned frames. The``extent``
    and ``period`` attributes are limited to the intersection of the
    ancestor stores. If the ancestor stores do not overlap (spatially or
    temporally), the elementwise block will be empty.
    """

    def __init__(self, *args):
        super(BaseElementwise, self).__init__(*args)
        # check the timedelta so that an error is raised if incompatible
        if len(self._sources) > 1:
            self.timedelta  # NOQA

    @property
    def _sources(self):
        # go through the arguments and list the blocks
        return [arg for arg in self.args if isinstance(arg, RasterBlock)]

    def get_sources_and_requests(self, **request):
        start = request.get("start", None)
        stop = request.get("stop", None)

        if start is not None and stop is not None:
            # limit request to self.period so that resulting data is aligned
            period = self.period
            if period is not None:
                request["start"] = max(start, period[0])
                request["stop"] = min(stop, period[1])

        process_kwargs = {"dtype": self.dtype.name, "fillvalue": self.fillvalue}
        sources_and_requests = [(source, request) for source in self.args]

        return [(process_kwargs, None)] + sources_and_requests

    @property
    def timedelta(self):
        """ The period between timesteps in case of equidistant time. """
        if len(self._sources) == 1:
            return self._sources[0].timedelta

        timedeltas = [s.timedelta for s in self._sources]
        if any(timedelta is None for timedelta in timedeltas):
            return None  # None precedes

        # multiple timedeltas: assert that they are equal
        if not timedeltas[1:] == timedeltas[:-1]:
            raise ValueError(
                "Time resolutions of input rasters are not "
                "equal ({}).".format(timedeltas)
            )
        else:
            return timedeltas[0]

    @property
    def period(self):
        """ Return period datetime tuple. """
        if len(self._sources) == 1:
            return self._sources[0].period

        periods = [s.period for s in self._sources]
        if any(period is None for period in periods):
            return None  # None precedes

        # multiple periods: return the overlapping period
        start = max([p[0] for p in periods])
        stop = min([p[1] for p in periods])
        if stop < start:
            return None  # no overlap
        else:
            return start, stop

    @property
    def extent(self):
        """ Boundingbox of contents in WGS84 projection. """
        if len(self._sources) == 1:
            return self._sources[0].extent

        extents = [s.extent for s in self._sources]
        if any(extent is None for extent in extents):
            return None  # None precedes

        # multiple extents: return the overlapping box
        x1 = max([e[0] for e in extents])
        y1 = max([e[1] for e in extents])
        x2 = min([e[2] for e in extents])
        y2 = min([e[3] for e in extents])
        if x2 <= x1 or y2 <= y1:
            return None  # no overlap
        else:
            return x1, y1, x2, y2

    @property
    def dtype(self):
        dtype = np.result_type(*self.args)
        if np.issubdtype(dtype, np.integer) or dtype == np.bool:
            # use at least int32
            return np.result_type(dtype, np.int32)
        elif np.issubdtype(dtype, np.floating):
            # use at least float32
            return np.result_type(dtype, np.float32)
        else:
            return dtype

    @property
    def fillvalue(self):
        dtype = self.dtype
        if dtype == np.bool:
            return
        else:
            return get_dtype_max(dtype)

    @property
    def geometry(self):
        """Intersection of geometries in the projection of the first store
        geometry. """
        geometries = [x.geometry for x in self._sources]
        if any(x is None for x in geometries):
            return
        if len(geometries) == 1:
            return geometries[0]
        result = geometries[0]
        sr = result.GetSpatialReference()
        for geometry in geometries[1:]:
            if not geometry.GetSpatialReference().IsSame(sr):
                geometry = geometry.Clone()
                geometry.TransformTo(sr)
            result = result.Intersection(geometry)
        if result.GetArea() == 0.0:
            return
        return result

    @property
    def projection(self):
        """Projection of the data if they match, else None"""
        projection = self._sources[0].projection
        if projection is None:
            return
        for arg in self._sources[1:]:
            if projection != arg.projection:
                return
        return projection

    @property
    def geo_transform(self):
        geo_transform = self._sources[0].geo_transform
        if geo_transform is None:
            return
        geo_transform = GeoTransform(geo_transform)
        for arg in self._sources[1:]:
            other = arg.geo_transform
            if other is None or not geo_transform.aligns_with(other):
                return
        return geo_transform


class BaseMath(BaseElementwise):
    """A block that applies basic math to two store-like objects."""

    def __init__(self, a, b):
        for x in (a, b):
            if not isinstance(x, (RasterBlock, np.ndarray, float, int)):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(BaseMath, self).__init__(a, b)


class BaseComparison(BaseMath):
    """A block that applies basic comparisons to two store-like objects."""

    @property
    def dtype(self):
        return np.dtype("bool")


class BaseLogic(BaseElementwise):
    """A block that applies basic logical operations to two store-like objects.
    """

    def __init__(self, a, b):
        for x in (a, b):
            if isinstance(x, (RasterBlock, np.ndarray)):
                dtype = x.dtype
                if dtype != np.dtype("bool"):
                    raise TypeError("inputs must have boolean dtypes")
            elif not isinstance(x, bool):
                raise TypeError("'{}' object is not allowed".format(type(x)))
        super(BaseLogic, self).__init__(a, b)

    @property
    def dtype(self):
        return np.dtype("bool")

    @property
    def fillvalue(self):
        return None


def wrap_math_process_func(func):
    """This enables normal math functions to operate only on the data values.
    Nodata is propagated. In case of comparison operators, nodata becomes
    False, except for NotEqual, where it becomes True.

    'meta' and 'time' fields are propagated from the first source. """

    @wraps(func)  # propagates the name and docstring
    def math_process_func(process_kwargs, *args):
        compute_args = []  # the args for the math operation
        # perform the nodata masking manually, as numpy maskedarrays are slow
        nodata_mask = None

        dtype = process_kwargs["dtype"]
        fillvalue = process_kwargs["fillvalue"]

        for data in args:
            if data is None:
                return None
            if not isinstance(data, dict):
                compute_args.append(data)
            elif "time" in data or "meta" in data:
                # return the time / meta right away. assumes there are no
                # mixed requests and that time is aligned
                return data
            elif "values" in data:
                compute_args.append(data["values"])
                # update the nodata mask
                if data["values"].dtype == np.dtype("bool"):
                    continue  # boolean data does not contain nodata values.
                if "no_data_value" not in data:
                    continue
                _nodata_mask = data["values"] == data["no_data_value"]
                if nodata_mask is None:
                    nodata_mask = _nodata_mask
                else:
                    nodata_mask |= _nodata_mask
            else:
                raise TypeError("Cannot apply math function to value {}".format(data))

        if dtype == np.dtype("bool"):
            no_data_value = None
            if func is np.not_equal:
                fillvalue = True
            else:
                fillvalue = False
            func_kwargs = {}
        else:
            func_kwargs = {"dtype": dtype}
            no_data_value = fillvalue

        with np.errstate(all="ignore"):  # suppresses warnings
            result_values = func(*compute_args, **func_kwargs)

        if nodata_mask is not None:
            result_values[nodata_mask] = fillvalue
        return {"no_data_value": no_data_value, "values": result_values}

    return math_process_func


class Add(BaseMath):
    """
    Add two rasters together or add a constant value to a raster.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Addition term a
      b (RasterBlock, number): Addition term b

    Returns:
      RasterBlock containing the result of the addition.
	"""

    process = staticmethod(wrap_math_process_func(np.add))


class Subtract(BaseMath):
    """
    Subtract two rasters or subtract a constant value from a raster

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Term be subtracted from
      b (RasterBlock, number): Term to be subtracted

    Returns:
      RasterBlock containing the result of the function subtract.
    """

    process = staticmethod(wrap_math_process_func(np.subtract))


class Multiply(BaseMath):
    """
    Multiply two rasters or multiply a raster by a constant value.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Multiplication factor a
      b (RasterBlock, number): Multiplication factor b
     
    Returns:
      RasterBlock containing the result of the multiplication.
    """

    process = staticmethod(wrap_math_process_func(np.multiply))


class Divide(BaseMath):
    """
    Divide two rasters or divide a raster by a constant value.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Numerator
      b (RasterBlock, number): Denominator
     
    Returns:
      RasterBlock containing the result of the division.
    """

    process = staticmethod(wrap_math_process_func(np.divide))

    @property
    def dtype(self):
        # use at least float32
        return np.result_type(np.float32, *self.args)


class Power(BaseMath):
    """
    Exponential function with either a raster and a number or two rasters.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Base
      b (RasterBlock, number): Exponent
      
    Returns:
      RasterBlock containing the result of the exponential function. 
    """

    process = staticmethod(wrap_math_process_func(np.power))

    def __init__(self, a, b):
        # cast negative integer exponents to float, because if a is also of
        # integer type, power will raise a ValueError
        if isinstance(b, int) and b < 0:
            b = float(b)
        super(Power, self).__init__(a, b)


class Equal(BaseComparison):
    """
    Compares the values of two rasters and returns True for equal elements.

    This operation can be used to compare two rasters or to compare a raster
    with a static value. Note that 'no data' is not equal to 'no data':
    False is returned if any of the two terms is 'no data'.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Comparison term a
      b (RasterBlock, number): Comparison term b
      
    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.equal))


class NotEqual(BaseComparison):
    """
    Compares the values of two rasters and returns False for equal elements.

    This operation can be used to compare two rasters or to compare a raster
    with a static value. Note that 'no data' is not equal to 'no data':
    True is returned if any of the two terms is 'no data'.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Comparison term a
      b (RasterBlock, number): Comparison term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.not_equal))


class Greater(BaseComparison):
    """
    Compares the values of two rasters and returns True if an element in the
    first term is greater.

    This operation can be used to compare two rasters or to compare a raster
    with a static value. Note that False is returned if any of the two terms is
    'no data'.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.

    Args:
      a (RasterBlock, number): Comparison term a
      b (RasterBlock, number): Comparison term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.greater))


class GreaterEqual(BaseComparison):
    """"
    Compares the values of two rasters and returns True if an element in the
    first term is greater or equal.

    This operation can be used to compare two rasters or to compare a raster
    with a static value. Note that False is returned if any of the two terms is
    'no data'.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.
    
    Args:
      a (RasterBlock, number): Comparison term a
      b (RasterBlock, number): Comparison term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.greater_equal))


class Less(BaseComparison):
    """
    Compares the values of two rasters and returns True if an element in the
    first term is less.

    This operation can be used to compare two rasters or to compare a raster
    with a static value. Note that False is returned if any of the two terms is
    'no data'.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.
    
    Args:
      a (RasterBlock, number): Comparison term a
      b (RasterBlock, number): Comparison term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.less))


class LessEqual(BaseComparison):
    """
    Compares the values of two rasters and returns True if an element in the
    first term is less or equal.

    This operation can be used to compare two rasters or to compare a raster
    with a static value. Note that False is returned if any of the two terms is
    'no data'.

    Either one or both of the inputs should be a RasterBlock. In case of
    two raster inputs the temporal properties of the rasters should be equal,
    however spatial properties can be different.
    
    Args:
      a (RasterBlock, number): Comparison term a
      b (RasterBlock, number): Comparison term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.less_equal))


class Invert(BaseSingle):
    """
    Logically invert a raster (swap True and False).
    
    Takes a single input raster containing boolean values and outputs a boolean
    raster with the same spatial and temportal properties.

    Args:
      x (RasterBlock): Boolean raster with values to invert

    Returns:
      RasterBlock with boolean values opposite to the input raster. 
    """

    def __init__(self, x):
        super(Invert, self).__init__(x)
        if x.dtype != np.dtype("bool"):
            raise TypeError("input block must have boolean dtype")

    @staticmethod
    def process(data):
        if "values" in data:
            return {"values": ~data["values"], "no_data_value": None}
        else:
            return data

    @property
    def dtype(self):
        return np.dtype("bool")


class IsData(BaseSingle):
    """
    Returns True where raster has data.
    
    Takes a single input raster and outputs a boolean raster with the same
    spatial and temporal properties.

    Args:
      store (RasterBlock): Input raster

    Returns:
      RasterBlock with boolean values. 
    """

    def __init__(self, store):
        if store.dtype == np.dtype("bool"):
            raise TypeError("input block must not have boolean dtype")
        super(IsData, self).__init__(store)

    @staticmethod
    def process(data):
        if data is None or "values" not in data:
            return data
        values = data["values"]
        no_data_value = data["no_data_value"]
        return {"values": values != no_data_value, "no_data_value": None}

    @property
    def dtype(self):
        return np.dtype("bool")

    @property
    def fillvalue(self):
        return None


class IsNoData(IsData):
    """
    Returns True where raster has no data.

    Takes a single input raster and outputs a boolean raster with the same
    spatial and temporal properties.

    Args:
      store (RasterBlock): Input raster

    Returns:
      RasterBlock with boolean values. 
    """

    @staticmethod
    def process(data):
        if data is None or "values" not in data:
            return data
        values = data["values"]
        no_data_value = data["no_data_value"]
        return {"values": values == no_data_value, "no_data_value": None}


class And(BaseLogic):
    """
    Returns True where both inputs are True.

    Either one or both of the inputs should be a boolean RasterBlock. In case
    of two raster inputs the temporal properties of the rasters should be
    equal, however spatial properties can be different.
    
    Args:
      a (RasterBlock, boolean): Logical term a
      b (RasterBlock, boolean): Logical term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.logical_and))


class Or(BaseLogic):
    """
    Returns True where any of inputs is True.

    Either one or both of the inputs should be a boolean RasterBlock. In case
    of two raster inputs the temporal properties of the rasters should be
    equal, however spatial properties can be different.

    Args:
      a (RasterBlock, boolean): Logical term a
      b (RasterBlock, boolean): Logical term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.logical_or))


class Xor(BaseLogic):
    """
    Exclusive or: returns True where exactly one of the inputs is True.

    Where both inputs are True, False is returned.

    Either one or both of the inputs should be a boolean RasterBlock. In case
    of two raster inputs the temporal properties of the rasters should be
    equal, however spatial properties can be different.

    Args:
      a (RasterBlock, boolean): Logical term a
      b (RasterBlock, boolean): Logical term b

    Returns:
      RasterBlock containing boolean values
    """

    process = staticmethod(wrap_math_process_func(np.logical_xor))


class FillNoData(BaseElementwise):
    """
    Combines multiple rasters filling 'no data' values.

    Values at equal timesteps in the contributing rasters are considered
    starting with the leftmost input raster. Therefore, values from rasters
    that are more 'to the right' are shown in the result. 'no data' values are
    transparent and will show data of rasters more 'to the left'.

    The temporal properties of the rasters should be equal, however spatial
    properties can be different.

    Args:
      *args (list of RasterBlocks): Rasters to be combined.
      
    Returns:
      RasterBlock that combines values from the inputs.
    """

    def __init__(self, *args):
        # TODO Expand this block so that it takes ndarrays and scalars.
        for arg in args:
            if not isinstance(arg, RasterBlock):
                raise TypeError("'{}' object is not allowed".format(type(arg)))
        super(FillNoData, self).__init__(*args)

    @staticmethod
    def process(process_kwargs, *args):
        """Combine data, filling in nodata values."""
        data_list = []
        no_data_values = []
        for data in args:
            if data is None:
                continue
            elif "time" in data or "meta" in data:
                # return the time / meta right away. assumes there are no
                # mixed requests and that time is aligned
                return data
            elif "values" in data and "no_data_value" in data:
                data_list.append(data["values"])
                no_data_values.append(data["no_data_value"])

        dtype = process_kwargs['dtype']
        fillvalue = get_dtype_max(dtype)

        if not data_list:
            return

        # initialize values array
        values = np.full(data_list[0].shape, fillvalue, dtype=dtype)

        # populate values array
        for data, no_data_value in zip(data_list, no_data_values):
            # index is where the source has data
            index = get_index(data, no_data_value)
            values[index] = data[index]

        return {"values": values, "no_data_value": fillvalue}



"""
Module containing the RasterBlock base classes.
"""
from datetime import datetime as Datetime

from dask_geomodeling import Block


class RasterBlock(Block):
    """ The base block for temporal rasters.

    All RasterBlocks must be derived from this base class and must implement
    the following attributes:

    - ``period``: a tuple of datetimes
    - ``timedelta``: a datetime.timedelta (or None if nonequidistant)
    - ``extent``: a tuple ``(x1, y1, x2, y2)``
    - ``dtype``: a numpy dtype object
    - ``fillvalue``: a number
    - ``geometry``: OGR Geometry
    - ``projection``: WKT string
    - ``geo_transform``: a tuple of 6 numbers

    These attributes are ``None`` if the raster is empty.

    A raster data request contains the following fields:

    - mode: values (``'vals'``), time (``'time'``) or metadata (``'meta'``)
    - bbox: bounding box ``(x1, y1, x2, y2)``
    - projection: wkt spatial reference
    - width: data width
    - height: data height
    - start: start date as naive UTC datetime
    - stop: stop date as naive UTC datetime

    The data response is ``None`` or a dictionary with the following fields:

    - (if mode was ``"vals"``) ``"values"``: a three dimensional numpy ndarray
      of shape ``(bands, height, width)``
    - (if mode was ``"vals"``) ``"no_data_value"``: a number that represents
      'no data'. If the ndarray is a boolean, there is no 'no data' value.
    - (if mode was ``"time"'"``) ``"time"``: a list of naive UTC datetimes
      corresponding to the time axis
    - (if mode was ``"meta"``) ``"meta"``: a list of metadata values
      corresponding to the time axis
    """

    DEFAULT_ORIGIN = Datetime(1970, 1, 1, 0, 0)

    def __len__(self):
        """ Return number of temporal bands. """
        # all empty
        try:
            start, stop = self.period
        except TypeError:
            return 0  # period is None

        if start == stop:
            return 1

        timedelta = self.timedelta
        if timedelta is None:
            # hard way, since bands are not aligned
            return len(self.get_data(mode="time", start=start, stop=stop)["time"])

        # bands are aligned, just divide the period length by the delta
        period_seconds = (stop - start).total_seconds()
        delta_seconds = timedelta.total_seconds()
        return int(period_seconds / delta_seconds) + 1

    def __add__(self, other):
        from . import Add

        return Add(self, other)

    def __mul__(self, other):
        from . import Multiply

        return Multiply(self, other)

    def __neg__(self):
        from . import Multiply

        return Multiply(self, -1)

    def __sub__(self, other):
        from . import Subtract

        return Subtract(self, other)

    def __truediv__(self, other):
        from . import Divide

        return Divide(self, other)

    def __pow__(self, other):
        from . import Power

        return Power(self, other)

    def __eq__(self, other):
        from . import Equal

        return Equal(self, other)

    def __ne__(self, other):
        from . import NotEqual

        return NotEqual(self, other)

    def __gt__(self, other):
        from . import Greater

        return Greater(self, other)

    def __ge__(self, other):
        from . import GreaterEqual

        return GreaterEqual(self, other)

    def __lt__(self, other):
        from . import Less

        return Less(self, other)

    def __le__(self, other):
        from . import LessEqual

        return LessEqual(self, other)

    def __invert__(self):
        from . import Invert

        return Invert(self)

    def __and__(self, other):
        from . import And

        return And(self, other)

    def __or__(self, other):
        from . import Or

        return Or(self, other)

    def __xor__(self, other):
        from . import Xor

        return Xor(self, other)


class BaseSingle(RasterBlock):
    """
    Baseclass for all raster blocks that adjust a single raster
    """

    def __init__(self, store, *args):
        if not isinstance(store, RasterBlock):
            raise TypeError("'{}' object is not allowed".format(type(store)))
        super(BaseSingle, self).__init__(store, *args)

    @property
    def store(self):
        return self.args[0]

    def __len__(self):
        return len(self.store)

    @property
    def extent(self):
        return self.store.extent

    @property
    def period(self):
        return self.store.period

    @property
    def timedelta(self):
        return self.store.timedelta

    @property
    def dtype(self):
        return self.store.dtype

    @property
    def fillvalue(self):
        return self.store.fillvalue

    @property
    def geometry(self):
        return self.store.geometry

    @property
    def projection(self):
        return self.store.projection

    @property
    def geo_transform(self):
        return self.store.geo_transform



"""
Module containing raster sources.
"""
import numpy as np

from osgeo import gdal, gdal_array

from datetime import datetime, timedelta

from dask_geomodeling import utils

from .base import RasterBlock

__all__ = ["MemorySource", "RasterFileSource"]


class MemorySource(RasterBlock):
    """A raster source that interfaces data from memory.

    Nodata values are supported, but when upsampling the data, these are
    assumed to be 0 biasing data edges towards 0.

    The raster pixel with its topleft corner at [x, y] will define ranges
    [x, x + dx) and (y - dy, y]. Here [dx, dy] denotes the (unsigned) pixel
    size. The topleft corner and top and left edges belong to a pixel.

    :param data: the pixel values
        this value will be transformed in a 3D array (t, y, x)
    :param no_data_value: the pixel value that designates 'no data'
    :param projection: the projection of the given pixel values
    :param pixel_size: the size of one pixel (in units given by projection)
        if x and y pixel sizes differ, provide them in (x, y) order
    :param pixel_origin: the location (x, y) of pixel with index (0, 0)
    :param time_first: the timestamp of the first frame in data (in
        milliseconds since 1-1-1970)
    :param time_delta: the difference between two consecutive frames (in ms)
    :param metadata: a list of metadata corresponding to the input frames

    :type data: number or ndarray
    :type no_data_value: number
    :type projection: str
    :type pixel_size: float or length-2 iterable of floats
    :type pixel_origin: length-2 iterable of floats
    :type time_first: integer or naive datetime
    :type time_delta: integer or timedelta or NoneType
    :type metadata: list or NoneType
    """

    def __init__(
        self,
        data,
        no_data_value,
        projection,
        pixel_size,
        pixel_origin,
        time_first=0,
        time_delta=None,
        metadata=None,
    ):
        data = np.atleast_3d(data)
        no_data_value = data.dtype.type(no_data_value)
        projection = utils.get_epsg_or_wkt(projection)
        if not hasattr(pixel_size, "__iter__"):
            pixel_size = [pixel_size] * 2
        else:
            pixel_size = list(pixel_size)
            if len(pixel_size) != 2:
                raise ValueError("pixel_size should have length 2")
        pixel_size = [float(x) for x in pixel_size]
        pixel_origin = [float(x) for x in pixel_origin]
        if len(pixel_origin) != 2:
            raise ValueError("pixel_origin should have length 2")
        if isinstance(time_first, datetime):
            time_first = utils.dt_to_ms(time_first)
        else:
            time_first = int(time_first)
        if isinstance(time_delta, timedelta):
            time_delta = int(time_delta.total_seconds() * 1000)
        elif time_delta is None:
            if data.shape[0] > 1:
                raise ValueError("time_delta is required for temporal data")
        else:
            time_delta = int(time_delta)
        if metadata is not None:
            metadata = list(metadata)
            if len(metadata) != data.shape[0]:
                raise ValueError("Metadata length should match data length")
        super().__init__(
            data,
            no_data_value,
            projection,
            pixel_size,
            pixel_origin,
            time_first,
            time_delta,
            metadata,
        )

    @property
    def data(self):
        return self.args[0]

    @property
    def no_data_value(self):
        return self.args[1]

    @property
    def projection(self):
        return self.args[2]

    @property
    def pixel_size(self):
        return self.args[3]

    @property
    def pixel_origin(self):
        return self.args[4]

    @property
    def time_first(self):
        return self.args[5]

    @property
    def time_delta(self):
        return self.args[6]

    @property
    def metadata(self):
        return self.args[7]

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def fillvalue(self):
        return self.no_data_value

    @property
    def geo_transform(self):
        p, q = self.pixel_origin
        a, d = self.pixel_size
        return utils.GeoTransform((p, a, 0, q, 0, -d))

    def _get_extent(self):
        if not self.data.size:
            return
        bbox = self.geo_transform.get_bbox((0, 0), self.data.shape[1:])
        return utils.Extent(bbox, utils.get_sr(self.projection))

    @property
    def extent(self):
        extent = self._get_extent()
        if extent is None:
            return
        return extent.transformed(utils.EPSG4326).bbox

    @property
    def geometry(self):
        extent = self._get_extent()
        if extent is None:
            return
        return extent.as_geometry()

    def __len__(self):
        return self.data.shape[0]

    @property
    def period(self):
        if len(self) == 0:
            return
        elif len(self) == 1:
            return (datetime.utcfromtimestamp(self.time_first / 1000),) * 2
        else:
            first = datetime.utcfromtimestamp(self.time_first / 1000)
            last = first + (len(self) - 1) * self.timedelta
            return first, last

    @property
    def timedelta(self):
        if len(self) <= 1:
            return None
        return timedelta(milliseconds=self.time_delta)

    def get_sources_and_requests(self, **request):
        mode = request["mode"]

        if (
            mode == "vals"
            and request.get("projection").upper() != self.projection.upper()
        ):
            raise RuntimeError("This source block cannot reproject data")
        elif mode == "meta" and self.metadata is None:
            return [({"mode": "empty_meta"}, None)]

        # interpret start and stop request parameters
        start, stop, first_i, last_i = utils.snap_start_stop(
            request.get("start"),
            request.get("stop"),
            datetime.utcfromtimestamp(self.time_first / 1000),
            self.timedelta,
            len(self),
        )
        if start is None:
            return [({"mode": "empty_" + request["mode"]}, None)]

        # set the process kwargs depending on the mode
        if mode == "vals":
            process_kwargs = {
                "mode": "vals",
                "data": self.data[first_i : last_i + 1],
                "no_data_value": self.no_data_value,
                "bbox": request["bbox"],
                "width": request["width"],
                "height": request["height"],
                "geo_transform": self.geo_transform,
            }
        elif mode == "meta":
            # metadata can't be None at this point
            process_kwargs = {
                "mode": "meta",
                "metadata": self.metadata[first_i : last_i + 1],
            }
        elif mode == "time":
            process_kwargs = {
                "mode": "time",
                "start": start,
                "delta": self.timedelta or timedelta(0),
                "length": last_i - first_i + 1,
            }
        else:
            raise RuntimeError("Unknown mode '{}'".format(mode))
        return [(process_kwargs, None)]

    @staticmethod
    def process(process_kwargs):
        mode = process_kwargs["mode"]

        # handle empty results
        if mode == "empty_vals":
            return
        elif mode == "empty_time":
            return {"time": []}
        elif mode == "empty_meta":
            return {"meta": []}

        # handle time requests
        if mode == "time":
            start = process_kwargs["start"]
            length = process_kwargs["length"]
            delta = process_kwargs["delta"]
            return {"time": [start + i * delta for i in range(length)]}
        elif mode == "meta":
            return {"meta": process_kwargs["metadata"]}

        # handle 'vals' requests
        data = process_kwargs["data"]
        no_data_value = process_kwargs["no_data_value"]
        bbox = process_kwargs["bbox"]
        width = process_kwargs["width"]
        height = process_kwargs["height"]
        gt = utils.GeoTransform(process_kwargs["geo_transform"])

        # return an empty array if 0-sized data was requested
        if width == 0 or height == 0:
            return np.empty((data.shape[0], height, width), dtype=data.dtype)

        # transform the requested bounding box to indices into the array
        shape = data.shape
        ranges, padding = gt.get_array_ranges(bbox, shape)
        result = data[:, slice(*ranges[0]), slice(*ranges[1])]

        # pad the data to the shape given by the index
        if padding is not None:
            padding = ((0, 0),) + padding  # for the time axis
            result = np.pad(result, padding, "constant", constant_values=no_data_value)

        # zoom to the desired height and width
        result = utils.zoom_raster(result, no_data_value, height, width)

        # fill nan values if they popped up
        result[~np.isfinite(result)] = no_data_value
        return {"values": result, "no_data_value": no_data_value}


class RasterFileSource(RasterBlock):
    """A raster source that interfaces data from a file path.

    The value at raster cell with its topleft corner at [x, y] is assumed to
    define a value for ranges [x, x + dx) and (y - dy, y]. Here [dx, dy]
    denotes the (unsigned) pixel size. The topleft corner and top and
    left edges belong to a pixel.

    :param url: the path to the file. File paths have to be contained inside
      the current root setting. Relative paths are interpreted relative to this
      setting (but internally stored as absolute paths).
    :param time_first: the timestamp of the first frame in data (in
        milliseconds since 1-1-1970), defaults to 1-1-1970
    :param time_delta: the difference between two consecutive frames (in ms),
        defaults to 5 minutes

    :type url: str
    :type time_first: integer or datetime
    :type time_delta: integer or timedelta

    The global root path can be adapted as follows:
      >>> from dask import config
      >>> config.set({"geomodeling.root": "/my/data/path"})

    Note that this object keeps a file handle open. If you need to close the
    file handle, call block.close_dataset (or dereference the whole object).
    """

    def __init__(self, url, time_first=0, time_delta=300000):
        url = utils.safe_file_url(url)
        if isinstance(time_first, datetime):
            time_first = utils.dt_to_ms(time_first)
        else:
            time_first = int(time_first)
        if isinstance(time_delta, timedelta):
            time_delta = int(time_delta.total_seconds() * 1000)
        else:
            time_delta = int(time_delta)
        super().__init__(url, time_first, time_delta)

    @property
    def url(self):
        return self.args[0]

    @property
    def time_first(self):
        return self.args[1]

    @property
    def time_delta(self):
        return self.args[2]

    @property
    def gdal_dataset(self):
        try:
            return self._gdal_dataset
        except AttributeError:
            path = utils.safe_abspath(self.url)
            self._gdal_dataset = gdal.Open(path)
            return self._gdal_dataset

    def close_dataset(self):
        if hasattr(self, "_gdal_dataset"):
            self._gdal_dataset = None

    @property
    def projection(self):
        return utils.get_epsg_or_wkt(self.gdal_dataset.GetProjection())

    @property
    def dtype(self):
        first_band = self.gdal_dataset.GetRasterBand(1)
        return gdal_array.GDALTypeCodeToNumericTypeCode(first_band.DataType)

    @property
    def fillvalue(self):
        first_band = self.gdal_dataset.GetRasterBand(1)
        return self.dtype(first_band.GetNoDataValue())

    @property
    def geo_transform(self):
        return utils.GeoTransform(self.gdal_dataset.GetGeoTransform())

    def _get_extent(self):
        bbox = self.geo_transform.get_bbox(
            (0, 0), (self.gdal_dataset.RasterYSize, self.gdal_dataset.RasterXSize)
        )
        return utils.Extent(bbox, utils.get_sr(self.projection))

    @property
    def extent(self):
        extent_epsg4326 = self._get_extent().transformed(utils.EPSG4326)
        return extent_epsg4326.bbox

    @property
    def geometry(self):
        return self._get_extent().as_geometry()

    def __len__(self):
        return self.gdal_dataset.RasterCount

    @property
    def period(self):
        if len(self) == 0:
            return
        elif len(self) == 1:
            return (datetime.utcfromtimestamp(self.time_first / 1000),) * 2
        else:
            first = datetime.utcfromtimestamp(self.time_first / 1000)
            last = first + (len(self) - 1) * self.timedelta
            return first, last

    @property
    def timedelta(self):
        if len(self) <= 1:
            return None
        return timedelta(milliseconds=self.time_delta)

    def get_sources_and_requests(self, **request):
        mode = request["mode"]

        if (
            mode == "vals"
            and request.get("projection").upper() != self.projection.upper()
        ):
            raise RuntimeError("This source block cannot reproject data")

        # interpret start and stop request parameters
        start, stop, first_i, last_i = utils.snap_start_stop(
            request.get("start"),
            request.get("stop"),
            datetime.utcfromtimestamp(self.time_first / 1000),
            self.timedelta,
            len(self),
        )
        if start is None:
            return [({"mode": "empty_" + request["mode"]}, None)]

        # set the process kwargs depending on the mode
        if mode == "vals":
            process_kwargs = {
                "mode": "vals",
                "url": self.url,
                "bbox": request["bbox"],
                "width": request["width"],
                "height": request["height"],
                "first_band": last_i,
                "last_band": first_i,
                "dtype": self.dtype,
                "fillvalue": self.fillvalue,
            }
        elif mode == "meta":
            # metadata can't be None at this point
            process_kwargs = {
                "mode": "meta",
                "url": self.url,
                "first_band": last_i,
                "last_band": first_i,
            }
        elif mode == "time":
            process_kwargs = {
                "mode": "time",
                "start": start,
                "delta": self.timedelta or timedelta(0),
                "length": last_i - first_i + 1,
            }
        else:
            raise RuntimeError("Unknown mode '{}'".format(mode))
        return [(process_kwargs, None)]

    @staticmethod
    def process(process_kwargs):
        mode = process_kwargs["mode"]

        # handle empty requests
        if mode == "empty_vals":
            return
        elif mode == "empty_time":
            return {"time": []}
        elif mode == "empty_meta":
            return {"meta": []}

        # handle time requests
        if mode == "time":
            start = process_kwargs["start"]
            length = process_kwargs["length"]
            delta = process_kwargs["delta"]
            return {"time": [start + i * delta for i in range(length)]}

        # open the dataset
        url = process_kwargs["url"]
        path = utils.safe_abspath(url)
        dataset = gdal.Open(path)
        first_band = process_kwargs["first_band"]
        last_band = process_kwargs["last_band"]

        # handle meta requests
        if mode == "meta":
            return {
                "meta": [
                    dataset.GetRasterBand(i + 1).GetMetadata_Dict()
                    for i in range(first_band, last_band + 1)
                ]
            }

        # handle 'vals' requests
        dtype = process_kwargs["dtype"]
        no_data_value = process_kwargs["fillvalue"]
        bbox = process_kwargs["bbox"]
        width = process_kwargs["width"]
        height = process_kwargs["height"]
        length = last_band - first_band + 1

        # return an empty array if 0-sized data was requested
        if width == 0 or height == 0:
            return np.empty((length, height, width), dtype=dtype)

        # transform the requested bounding box to indices into the array
        shape = dataset.RasterCount, dataset.RasterYSize, dataset.RasterXSize
        gt = utils.GeoTransform(dataset.GetGeoTransform())
        ranges, padding = gt.get_array_ranges(bbox, shape)
        read_shape = [rng[1] - rng[0] for rng in ranges]

        # return nodata immediately for empty
        if any([x <= 0 for x in read_shape]):
            result = np.full(
                shape=(length, height, width), fill_value=no_data_value, dtype=dtype
            )
            return {"values": result, "no_data_value": no_data_value}

        # read arrays from file
        result = np.empty([length] + read_shape, dtype=dtype)
        for k in range(length):
            band = dataset.GetRasterBand(first_band + k + 1)
            result[k] = band.ReadAsArray(
                int(ranges[1][0]),
                int(ranges[0][0]),
                int(read_shape[1]),
                int(read_shape[0]),
            )

        # pad the data to the shape given by the index
        if padding is not None:
            padding = ((0, 0),) + padding  # for the time axis
            result = np.pad(result, padding, "constant", constant_values=no_data_value)

        # zoom to the desired height and width
        result = utils.zoom_raster(result, no_data_value, height, width)

        # fill nan values if they popped up
        result[~np.isfinite(result)] = no_data_value
        return {"values": result, "no_data_value": no_data_value}



from .base import RasterBlock  # NOQA
from .elemwise import *  # NOQA
from .combine import *  # NOQA
from .spatial import *  # NOQA
from .temporal import *  # NOQA
from .misc import *  # NOQA
from .sources import *  # NOQA



"""
Module containing the core graphs.
"""
import inspect
import sys
import json
import logging

from dask.base import tokenize, normalize_token
from dask.local import get_sync

from shapely.geometry.base import BaseGeometry
from datetime import datetime
from datetime import timedelta

logger = logging.getLogger(__name__)

__all__ = ["construct", "construct_multiple", "compute", "Block", "DummyBlock"]


def _construct_exc_callback(e, dumps):
    """Callback to be used as a 'pack_exception' kwarg in get_sync

    """
    key = inspect.currentframe().f_back.f_locals.get("key")
    e.args = ("{0}: {1}".format(key, str(e)),)
    raise e


def _reconstruct_token(key):
    """Reconstruct a token from a key in a graph ('SomeName_<token>')"""
    if len(key) < 34 or key[-33] != "_":
        return
    token = key[-32:]
    try:
        int(token, 16)
    except ValueError:
        return None
    return token.lower()


def compute(graph, name, *args, **kwargs):
    """Compute a graph ({name: [func, arg1, arg2, ...]}) using dask.get_sync
    """
    return get_sync(graph, [name])[0]


def construct(graph, name, validate=True):
    """Construct a Block with dependent Blocks from a graph and endpoint name.
    """
    return construct_multiple(graph, [name], validate)[0]


def construct_multiple(graph, names, validate=True):
    """Construct multiple Blocks from given graph and endpoint names.
    """
    # deserialize import paths where necessary and cast lists to tuples
    new_graph = {}
    for key, value in graph.items():
        cls = value[0]
        if isinstance(cls, str):
            cls = Block.from_import_path(cls)
        if not issubclass(cls, Block):
            raise TypeError("Cannot construct from object of type '{}'".format(cls))
        args = tuple(value[1:])
        if validate:
            new_graph[key] = (cls,) + args
        else:
            token = _reconstruct_token(key)
            if token is None:
                logger.warning(
                    "Construct received a key with an invalid name ('%s'),"
                    "while validation was turned off",
                    key,
                )
            new_graph[key] = (cls._init_no_validation, token) + args

    return get_sync(new_graph, names, pack_exception=_construct_exc_callback)


class Block(object):
    """ A class that generates dask-like compute graphs for given requests.

    Arguments (args) are always stored in ``self.args``.
    If a request is passed into the Block using the ``get_data`` or
    (the lazy version) ``get_compute_graph`` method, the Block figures out
    what args are actually necessary to evaluate the request, and what
    requests need to be sent to those args. This happens in the method
    ``get_sources_and_requests``.

    After the requests have been evaluated, the data comes back and is passed
    into the ``process`` method.
    """

    JSON_VERSION = 2

    @property
    def token(self):
        """Generates a unique and deterministic representation of this object
        """
        # the token is cached on this Block object
        try:
            return self._cached_token
        except AttributeError:
            pass
        klass_path = self.get_import_path()
        args = [arg.token if isinstance(arg, Block) else arg for arg in self.args]
        self._cached_token = tokenize(klass_path, *args)
        return self._cached_token

    @staticmethod  # must be a static method
    def process(data):
        """
        Overridden to modify data from sources in unlimited ways.

        Default implementation passes single-source unaltered data.
        """
        return data

    def __init__(self, *args):
        """The init should be overridden by subclasses. In each __init__:
         - Explicitly give a call signature and a corresponding docstring
         - Check the types of the provided args, so that exceptions are raised
           during pipeline construction.
         - Call super().__init__().
        """
        self.args = args

    @classmethod
    def _init_no_validation(cls, token, *args):
        """This constructs this block directly from its key and args
        without any validation and tokenization."""
        obj = cls.__new__(cls)
        obj.args = args
        if token:
            obj._cached_token = token
        return obj

    def get_sources_and_requests(self, **request):
        """Adapt the request and/or select the sources to be computed. The
        request is allowed to differ per source.

        This function should return an iterable of (source, request). For
        sources that are no Block instance, the request is ignored.

        Exceptions raised here will be raised before actual computation starts.
        (at .get_compute_graph(request)).
        """
        return ((source, request) for source in self.args)

    """Below are methods that should never be overridden by subclasses"""

    def get_data(self, **request):
        """Directly evaluate the request and return the data."""
        return compute(*self.get_compute_graph(**request))

    def get_compute_graph(self, cached_compute_graph=None, **request):
        """Lazy version of get_data, returns a compute graph dict, that
        can be evaluated with `compute` (or dask's get function).

        The dictionary has keys in the form ``name_token`` and values in the
        form ``tuple(process, *args)``, where ``args`` are the precise
        arguments that need to be passed to ``process``, with the exception
        that args may reference to other keys in the dictionary.
        """
        # generate a token from the specific request passed to the block
        # NB generates a random hash if the request cannot be tokenized
        token = tokenize([self.token, request])
        name = "{}_{}".format(self.__class__.__name__.lower(), token)
        graph = cached_compute_graph or dict()

        if name in graph:
            return graph, name

        args = [self.process]
        for source, req in self.get_sources_and_requests(**request):
            if isinstance(source, Block) and req is not None:
                graph, compute_name = source.get_compute_graph(
                    cached_compute_graph=graph, **req
                )
                args.append(compute_name)
            else:
                args.append(source)

        graph[name] = tuple(args)
        return graph, name

    def get_graph(self, serialize=False):
        """Generate a graph that defines this Block and its dependencies
        in a dictionary.

        The dictionary has keys in the form ``name_token`` and values in the
        form ``tuple(Block class, *args)``, where ``args`` are the precise
        arguments that were used to construct the Block, with the exception
        that args may also reference other keys in the dictionary.

        If serialize == True, the Block classes will be replaced by their
        corresponding import paths.
        """
        if serialize:
            args = [self.get_import_path()]
        else:
            args = [self.__class__]
        graph = dict()
        for arg in self.args:
            if isinstance(arg, Block):
                sub_graph, sub_name = arg.get_graph(serialize=serialize)
                graph.update(sub_graph)
                args.append(sub_name)
            else:
                args.append(arg)
        name = self.name
        graph[name] = args
        return graph, name

    @property
    def name(self):
        return "{}_{}".format(self.__class__.__name__, self.token)

    def __reduce__(self):
        """Serialize the object (pattern: callable, args). Construct is called
        with the arguments (graph, name) to reconstruct this block and its
        dependencies. Validation is skipped by adding False to the args."""
        return construct, self.get_graph() + (False,)

    @classmethod
    def get_import_path(cls):
        """Serialize the Block by returning its import path."""
        name = cls.__name__
        module = cls.__module__

        try:
            __import__(module)
            mod = sys.modules[module]
            klass = getattr(mod, name)
        except (ImportError, KeyError, AttributeError):
            raise Exception(
                "Can't serialize %r: it's not found as %s.%s" % (cls, module, name)
            )
        else:
            if klass is not cls:
                raise Exception(
                    "Can't serialize %r: it's not the same object as %s.%s"
                    % (cls, module, name)
                )

        return "{}.{}".format(module, name)

    @staticmethod
    def from_import_path(path):
        """Deserialize the Block by importing it from given path."""
        module, name = path.rsplit(".", 1)
        __import__(module)
        mod = sys.modules[module]
        klass = getattr(mod, name)
        if issubclass(klass, Block):
            return klass
        else:
            raise TypeError('"{}" is not valid Block.'.format(path))

    @classmethod
    def from_json(cls, val, **kwargs):
        """Construct a graph from a json stream."""
        return cls.deserialize(json.loads(val, **kwargs))

    def to_json(self, **kwargs):
        """Dump the graph to a json stream."""
        return json.dumps(self.serialize(), **kwargs)

    def serialize(self):
        """Serialize this block into a dict containing version, graph and name
        """
        graph, name = self.get_graph(serialize=True)
        return {"version": self.JSON_VERSION, "graph": graph, "name": name}

    @classmethod
    def deserialize(cls, val, validate=False):
        """
        Deserialize this block from a dict containing version, graph and name
        """
        # TODO Compare val['version'] with cls.JSON_VERSION
        return construct(val["graph"], val["name"], validate=validate)

    def __repr__(self):
        name = self.__class__.__name__
        return "{}({})".format(name, ", ".join([repr(x) for x in self.args]))


class DummyBlock(Block):
    """This dummy block pretends that it has the user-supplied name and token

    This is useful for partially evaluating block graphs for computing tokens.
    """

    def __init__(self, name):
        super().__init__(name)

    @property
    def token(self):
        return self.name.split("_")[1]

    @property
    def name(self):
        return self.args[0]


# Dask knows how to hash a couple of object types. If it doesn't know how to,
# it iterates over the MRO of the object until it ends up at 'object'. Then
# it checks for the __dask_tokenize__ slot, and if it does not exist, it
# generates a random hash.

# Tokenize shapely geometry by using its WKB representation.
@normalize_token.register(BaseGeometry)
def normalize_shapely_geometry(geometry):
    return geometry.wkb


# Tokenize datetime and timedeltas using their pickle handle
@normalize_token.register((datetime, timedelta))
def normalize_datetime(value):
    return hash(value)



import re
import math
import pytz
import os
import warnings
from distutils.version import LooseVersion
from functools import lru_cache
from itertools import repeat

from math import floor, log10

import numpy as np
import pandas as pd
from scipy import ndimage
from dask import config

from osgeo import gdal, ogr, osr, gdal_array
from shapely.geometry import box, Point
from shapely import wkb as shapely_wkb

import fiona
import fiona.crs
import geopandas


POLYGON = "POLYGON (({0} {1},{2} {1},{2} {3},{0} {3},{0} {1}))"

try:
    from fiona import Env as fiona_env  # NOQA
except ImportError:
    from fiona import drivers as fiona_env  # NOQA

try:
    GEOPANDAS_GTE_0_7_0 = LooseVersion(geopandas.__version__) >= LooseVersion("0.7.0")
except AttributeError:  # for doctests geopandas is mocked
    GEOPANDAS_GTE_0_7_0 = True

if GEOPANDAS_GTE_0_7_0:
    from pyproj import CRS


def get_index(values, no_data_value):
    """ Return an index to access for data values in values. """
    equal = np.isclose if values.dtype.kind == "f" else np.equal
    return np.logical_not(equal(values, no_data_value))


def get_dtype_max(dtype):
    """
    Return the maximum value for a dtype as a python scalar value.

    :param dtype: numpy dtype
    """
    d = np.dtype(dtype)
    if d.kind == "f":
        return np.finfo(d).max.item()
    return np.iinfo(d).max


def get_dtype_min(dtype):
    """
    Return the minimum value for a dtype as a python scalar value.

    :param dtype: numpy dtype
    """
    d = np.dtype(dtype)
    if d.kind == "f":
        return np.finfo(d).min.item()
    return np.iinfo(d).min


def get_uint_dtype(n):
    """Get the smallest uint dtype that accomodates 'n' values"""
    for dtype in ("u1", "u2", "u4", "u8"):
        if n - 1 <= np.iinfo(dtype).max:
            return np.dtype(dtype)
    raise ValueError("Too many values for uint dtype ({})".format(n))


def get_rounded_repr(obj, significant=4, fmt="{} (rounded)"):
    """
    http://stackoverflow.com/questions/3410976/
    how-to-round-a-number-to-significant-figures-in-python

    :param significant: number of significant digits
    :param fmt: template for rounded repr
    """
    digits = (
        -int(floor(log10(abs(n)))) + (significant - 1) if n else None for n in obj
    )
    rounded = obj.__class__(round(n, d) if n else n for n, d in zip(obj, digits))
    if obj == rounded:
        return repr(obj)
    return fmt.format(repr(rounded))


class Extent(object):
    """ Spatially aware extent. """

    def __init__(self, bbox, sr):
        self.bbox = bbox
        self.sr = sr

    def __repr__(self):
        return "<{}: {} / {}>".format(
            self.__class__.__name__,
            get_projection(self.sr),
            get_rounded_repr(self.bbox),
        )

    @property
    def width(self):
        return self.bbox[2] - self.bbox[0]

    @property
    def height(self):
        return self.bbox[3] - self.bbox[1]

    @classmethod
    def from_geometry(cls, geometry):
        """ Return Extent instance. """
        x1, x2, y1, y2 = geometry.GetEnvelope()
        bbox = x1, y1, x2, y2
        sr = geometry.GetSpatialReference()
        return cls(bbox=bbox, sr=sr)

    def as_geometry(self):
        """ Return ogr Geometry instance. """
        x1, y1, x2, y2 = self.bbox

        ring = ogr.Geometry(ogr.wkbLinearRing)
        ring.AddPoint_2D(x1, y1)
        ring.AddPoint_2D(x2, y1)
        ring.AddPoint_2D(x2, y2)
        ring.AddPoint_2D(x1, y2)
        ring.AddPoint_2D(x1, y1)
        polygon = ogr.Geometry(ogr.wkbPolygon)
        polygon.AddGeometry(ring)
        polygon.AssignSpatialReference(self.sr)
        return polygon

    def buffered(self, size):
        """ Return Extent instance. """
        x1, y1, x2, y2 = self.bbox
        buffered_bbox = x1 - size, y1 - size, x2 + size, y2 + size
        return self.__class__(bbox=buffered_bbox, sr=self.sr)

    def transformed(self, sr):
        geometry = self.as_geometry()
        geometry.TransformTo(sr)
        return Extent.from_geometry(geometry)


class GeoTransform(tuple):
    """
    Wrapper with handy methods for the GeoTransform tuple
    as used by the GDAL library.

    In the dask-geomodeling as well as in the GDAL library the geo_transform
    defines the transformation from array pixel indices to projected
    coordinates. A pair of projected coordinates (x, y) is calculated
    from a pair of array indices (i, j) as follows:

    p, a, b, q, c, d = self  # the geo_transform tuple

    x = p + a * j + b * i
    y = q + c * j + d * i

    or in a kind of vector notation:

    [x, y] = [p, q] + [[a, b], [c, d]] * [j, i]
    """

    @classmethod
    def from_bbox(cls, bbox, height, width):
        x1, y1, x2, y2 = bbox
        p, q = x1, y2
        b, c = 0, 0
        a = (x2 - x1) / width
        d = (y1 - y2) / height
        return cls((p, a, b, q, c, d))

    def __init__(self, tpl):
        if len(tpl) != 6:
            raise ValueError("GeoTransform expected an iterable of length 6")
        if not all(math.isclose(tpl[i], 0.0, abs_tol=1e-7) for i in (2, 4)):
            raise ValueError("Tilted geo_transforms are not supported")
        if any(math.isclose(tpl[i], 0.0, abs_tol=1e-7) for i in (1, 5)):
            raise ValueError("Pixel size should not be zero")

    def __repr__(self):
        return get_rounded_repr(tuple(self))

    @property
    def cell_area(self):
        p, a, b, q, c, d = self
        return abs(a * d - b * c)

    @property
    def origin(self):
        """Return the (x, y) coordinate of pixel with indices (0, 0)."""
        return self[0], self[3]

    @property
    def origin_normalized(self):
        """Return the (x, y) coordinate of the pixel closest to x = 0, y = 0.
        """
        return self[0] % self[1], self[3] % self[5]

    def get_inverse(self):
        """Return 2 x 2 matrix for the inverse of self"""
        _, a, b, _, c, d = self
        D = 1 / (a * d - b * c)
        return d * D, -b * D, -c * D, a * D

    def scale(self, x, y):
        """
        return scaled geo transform.

        :param x: Multiplication factor for the pixel width.
        :param y: Multiplication factor for the pixel height.

        Adjust the second, third, fifth and sixth elements of the geo
        transform so that the extent of the respective image is multiplied
        by scale.
        """
        p, a, b, q, c, d = self
        return self.__class__([p, a * x, b * x, q, c * y, d * y])

    def shift(self, origin):
        """
        Return shifted geo transform.

        :param origin: Integer pixel coordinates.

        Adjust the first and fourth element of the geo transform to match
        a subarray that starts at origin.
        """
        p, a, b, q, c, d = self
        i, j = origin
        return self.__class__([p + a * j + b * i, a, b, q + c * j + d * i, c, d])

    def get_indices(self, points):
        """
        Return pixel indices as a tuple of linear numpy arrays
        """
        # inverse transformation
        p, _, _, q, _, _ = self
        e, f, g, h = self.get_inverse()

        # calculate
        x, y = np.asarray(points).transpose()
        return (
            np.floor(g * (x - p) + h * (y - q)).astype(np.int64),
            np.floor(e * (x - p) + f * (y - q)).astype(np.int64),
        )

    def get_points(self, indices):
        """
        Return point coordinates as N x 2 numpy array.
        """
        p, a, b, q, c, d = self
        i, j = indices.transpose()
        points = np.empty(indices.shape)
        points[:, 0] = p + a * j + b * i
        points[:, 1] = q + c * j + d * i
        return points

    def get_bbox(self, offset, shape):
        """Return the bounding box coordinates defined by an offset and shape
        of an array"""
        p, a, b, q, c, d = self
        i, j = offset
        m, n = shape
        west = p + a * j + b * i
        north = q + c * j + d * i
        east = west + a * n + b * m
        south = north + c * n + d * m
        return west, south, east, north

    def get_indices_for_bbox(self, bbox):
        """Return (i1, i2), (j1, j2) array coordinates for given bbox."""
        x1, y1, x2, y2 = bbox

        # inverse transformation
        p, _, _, q, _, _ = self
        e, f, g, h = self.get_inverse()

        # apply to envelope corners
        x_index_1 = int(floor(e * (x1 - p) + f * (y2 - q)))
        y_index_1 = int(floor(g * (x1 - p) + h * (y2 - q)))
        x_index_2 = int(floor(e * (x2 - p) + f * (y1 - q)))
        y_index_2 = int(floor(g * (x2 - p) + h * (y1 - q)))

        # sort for flipped geo-transfoms
        x_pair = tuple(sorted((x_index_1, x_index_2)))
        y_pair = tuple(sorted((y_index_1, y_index_2)))
        return y_pair, x_pair

    def get_array_ranges(self, bbox, shape):
        """Compute how much to slice and pad an array with given shape to
        obtain `bbox`.

        :param bbox: tuple of (x1, y1, x2, y2)
        :param shape: tuple of 2 int

        :returns:
          - two [start, stop) ranges for the to array dimensions
          - two [before, after] paddings, or None if no padding is required
        """
        (i1, i2), (j1, j2) = self.get_indices_for_bbox(bbox)

        # increase the end index by one if necessary to deal with points
        if i1 == i2:
            i2 += 1
        if j1 == j2:
            j2 += 1

        # clip the indices to the array bounds for getting the data
        _i1, _i2 = np.clip([i1, i2], 0, shape[1])
        _j1, _j2 = np.clip([j1, j2], 0, shape[2])
        ranges = (_i1, _i2), (_j1, _j2)

        # pad the data to the shape given by the index
        padding_i = (i2 - i1, 0) if _i1 == _i2 else (_i1 - i1, i2 - _i2)
        padding_j = (j2 - j1, 0) if _j1 == _j2 else (_j1 - j1, j2 - _j2)
        padding = padding_i, padding_j
        if np.all(np.array(padding) <= 0):
            padding = None
        return ranges, padding

    def aligns_with(self, other):
        """Compares with other geotransform, checks if rasters are aligned."""
        if not isinstance(other, GeoTransform):
            other = GeoTransform(other)
        # compare resolutions
        if abs(self[1]) != abs(other[1]) or abs(self[5]) != abs(other[5]):
            return False
        return self.origin_normalized == other.origin_normalized


@lru_cache(32)  # least-recently-used cache of size 32
def get_sr(user_input):
    """ Return osr.SpatialReference for user input. """
    return osr.SpatialReference(osr.GetUserInputAsWKT(str(user_input)))


def get_crs(user_input):
    """Return CRS for user input.

    Args:
      user_input (str): a WKT, PROJ4, or EPSG:xxxx crs representation

    Returns:
      for geopandas >= 0.7: pyproj.CRS
      for geopandas < 0.7: dict
    """
    if GEOPANDAS_GTE_0_7_0:
        return CRS(user_input)
    wkt = osr.GetUserInputAsWKT(str(user_input))
    sr = osr.SpatialReference(wkt)
    key = str("GEOGCS") if sr.IsGeographic() else str("PROJCS")
    name = sr.GetAuthorityName(key)
    if name == "EPSG":
        # we can specify CRS in EPSG code which is more compatible with output
        # file types
        return fiona.crs.from_epsg(int(sr.GetAuthorityCode(key)))
    else:
        # we have to go through Proj4
        return fiona.crs.from_string(sr.ExportToProj4())


def crs_to_srs(crs):
    """
    Recover our own WKT definition of projections from a pyproj / fiona CRS

    Args:
      crs (pyproj.CRS or dict): what is returned from GeoDataFrame().crs

    Returns:
      string: a "EPSG:xxx" or WKT representation of the crs.
    """
    if crs is None:
        return
    elif isinstance(crs, dict):  # geopandas < 0.7
        proj4_str = fiona.crs.to_string(crs)
        return get_epsg_or_wkt(proj4_str)
    else:  # geopandas >= 0.7
        return crs.to_string()


def wkb_transform(wkb, src_sr, dst_sr):
    """
    Return a shapely geometry transformed from src_sr to dst_sr.

    :param wkb: wkb bytes
    :param src_sr: source osr SpatialReference
    :param dst_sr: destination osr SpatialReference
    """
    result = ogr.CreateGeometryFromWkb(wkb, src_sr)
    result.TransformTo(dst_sr)
    return result.ExportToWkb()


def shapely_transform(geometry, src_srs, dst_srs):
    """
    Return a shapely geometry transformed from src_srs to dst_srs.

    :param geometry: shapely geometry
    :param src_srs: source projection string
    :param dst_srs: destination projection string

    Note that we do not use geopandas for the transformation, because this is
    much slower than OGR.
    """
    return shapely_wkb.loads(
        wkb_transform(geometry.wkb, src_sr=get_sr(src_srs), dst_sr=get_sr(dst_srs))
    )


def geoseries_transform(df, src_srs, dst_srs):
    """
    Transform a GeoSeries to a different SRS. Returns a copy.

    :param df: GeoSeries to transform
    :param src_srs: source projection string
    :param dst_srs: destination projection string

    Note that we do not use .to_crs() for the transformation, because this is
    much slower than OGR. Also, we ignore the .crs property (but we do set it)
    """
    result = df.geometry.apply(shapely_transform, args=(src_srs, dst_srs))
    result.crs = get_crs(dst_srs)
    return result


def geodataframe_transform(df, src_srs, dst_srs):
    """
    Transform the geometry column of a GeoDataFrame to a different SRS.

    :param df: GeoDataFrame to transform (will be changed inplace)
    :param src_srs: source projection string
    :param dst_srs: destination projection string

    Note that we do not use .to_crs() for the transformation, because this is
    much slower than OGR. Also, we ignore the .crs property (but we do set it)
    """
    geoseries = geoseries_transform(df.geometry, src_srs, dst_srs)
    df.crs = geoseries.crs
    df.set_geometry(geoseries)
    return df


def transform_min_size(min_size, geometry, src_srs, dst_srs):
    """
    Return min_size converted from source crs to a target crs.

    :param min_size: float indicating the minimum size in the source crs
    :param geometry: shapely geometry of the location in the source crs
    :param src_srs: source projection string
    :param dst_srs: destination projection string

    Note that in order to guarantee the minimum size, the reverse operation may
    result in a larger value than the original minimum size.
    """
    source = geometry.centroid.buffer(min_size / 2)
    target = shapely_transform(source, src_srs=src_srs, dst_srs=dst_srs)
    x1, y1, x2, y2 = target.bounds
    return max(x2 - x1, y2 - y1)


def transform_extent(extent, src_srs, dst_srs):
    """
    Return extent tuple transformed from src_srs to dst_srs.

    :param extent: xmin, ymin, xmax, ymax
    :param src_srs: source projection string
    :param dst_srs: destination projection string
    """
    source = box(*extent)
    target = shapely_transform(source, src_srs=src_srs, dst_srs=dst_srs)
    return target.bounds


EPSG3857 = get_sr("EPSG:3857")
EPSG4326 = get_sr("EPSG:4326")


def get_projection(sr):
    """ Return simple userinput string for spatial reference, if any. """
    key = str("GEOGCS") if sr.IsGeographic() else str("PROJCS")
    return "{name}:{code}".format(
        name=sr.GetAuthorityName(key), code=sr.GetAuthorityCode(key)
    )


def get_epsg_or_wkt(text):
    """
    Return EPSG:<code> where possible, and WKT otherwise.

    :param text: Textual representation of a spatial reference system.
    """
    wkt = osr.GetUserInputAsWKT(str(text))
    sr = osr.SpatialReference(wkt)
    key = str("GEOGCS") if sr.IsGeographic() else str("PROJCS")
    name = sr.GetAuthorityName(key)
    if name is None:
        return wkt
    code = sr.GetAuthorityCode(key)
    return "{name}:{code}".format(name=name, code=code)


def get_footprint(size):
    """
    Return numpy array of booleans representing a circular footprint.

    :param size: diameter of the circle, coerced to uneven values.
    """
    s = size // 2 * 2 + 1
    o = (s - 1) // 2
    r = s / 2

    x, y = np.indices((s, s)) - o
    return (x ** 2 + y ** 2) < (r ** 2)


def create_dataset(array, geo_transform=None, projection=None, no_data_value=None):
    """
    Create and return a gdal dataset.

    :param array: A numpy array.
    :param geo_transform: 6-tuple of floats
    :param projection: wkt projection string
    :param no_data_value: integer or float

    This is the fastest way to get a gdal dataset from a numpy array, but
    keep a reference to the array around, or a segfault will occur. Also,
    don't forget to call FlushCache() on the dataset after any operation
    that affects the array.
    """
    # prepare dataset name pointing to array
    datapointer = array.ctypes.data
    bands, lines, pixels = array.shape
    datatypecode = gdal_array.NumericTypeCodeToGDALTypeCode(array.dtype.type)
    datatype = gdal.GetDataTypeName(datatypecode)
    bandoffset, lineoffset, pixeloffset = array.strides
    # if projection is wrong there will be a segfault in RasterizeLayer
    projection = osr.GetUserInputAsWKT(str(projection))

    dataset_name_template = (
        "MEM:::"
        "DATAPOINTER={datapointer},"
        "PIXELS={pixels},"
        "LINES={lines},"
        "BANDS={bands},"
        "DATATYPE={datatype},"
        "PIXELOFFSET={pixeloffset},"
        "LINEOFFSET={lineoffset},"
        "BANDOFFSET={bandoffset}"
    )
    dataset_name = dataset_name_template.format(
        datapointer=datapointer,
        pixels=pixels,
        lines=lines,
        bands=bands,
        datatype=datatype,
        pixeloffset=pixeloffset,
        lineoffset=lineoffset,
        bandoffset=bandoffset,
    )

    # access the array memory as gdal dataset
    dataset = gdal.Open(dataset_name, gdal.GA_Update)

    # set additional properties from kwargs
    if geo_transform is not None:
        dataset.SetGeoTransform(geo_transform)
    if projection is not None:
        dataset.SetProjection(projection)
    if no_data_value is not None:
        for i in range(len(array)):
            dataset.GetRasterBand(i + 1).SetNoDataValue(no_data_value)

    return dataset


class Dataset(object):
    """
    Usage:
        >>> with Dataset(array) as dataset:
        ...     # do gdal things.
    """

    def __init__(self, array, **kwargs):
        self.array = array
        self.dataset = create_dataset(array, **kwargs)

    def __enter__(self):
        return self.dataset

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.dataset.FlushCache()


def _finalize_rasterize_result(array, no_data_value):
    if array.dtype == np.uint8:  # cast to bool
        array = array.astype(np.bool)
        no_data_value = None
    return {"values": array, "no_data_value": no_data_value}


def rasterize_geoseries(geoseries, bbox, projection, height, width, values=None):
    """Transform a geoseries to a raster, optionally.

    :param geoseries: GeoSeries or None
    :param bbox: tuple of floats, (x1, y1, x2, y2)
    :param projection: wkt projection string
    :param height: int
    :param width: int
    :param values: Series or None, object containing values (int, float, bool)
      that will be burned into the raster. If None, the result will be a
      boolean array indicating geometries.

    :returns: dictionary containing
      - values: a 3D numpy array (time, y, x) of int32 / float64 / bool type
      - no_data_value

    It is assumed that all geometries intersect the requested bbox. If
    geoseries is None or empty, an array full of nodata will be returned.
    """
    # determine the dtype based on `values`
    if values is None or values.dtype == np.bool:
        dtype = np.uint8  # we cast to bool later as GDAL does not know bools
        no_data_value = 0  # False
        ogr_dtype = None
        if values is not None and geoseries is not None:
            geoseries = geoseries[values]  # values is a boolean mask
            values = None  # discard values
    elif str(values.dtype) == "category":
        # transform pandas Categorical dtype to normal dtype
        values = pd.Series(np.asarray(values), index=values.index)

    if values is not None:
        if np.issubdtype(values.dtype, np.floating):
            dtype = np.float64  # OGR only knows float64
            no_data_value = get_dtype_max(dtype)
            ogr_dtype = ogr.OFTReal
            if geoseries is not None:
                # filter out the inf and NaN values
                mask = np.isfinite(values)
                geoseries = geoseries[mask]
                values = values[mask]
        elif np.issubdtype(values.dtype, np.integer):
            dtype = np.int32  # OGR knows int32 and int64, but GDAL only int32
            no_data_value = get_dtype_max(dtype)
            ogr_dtype = ogr.OFTInteger
        else:
            raise TypeError(
                "Unsupported values dtype to rasterize: '{}'".format(values.dtype)
            )

    # initialize the array
    array = np.full((1, height, width), no_data_value, dtype=dtype)

    # if there are no features, return directly
    if geoseries is None or len(geoseries) == 0:
        return _finalize_rasterize_result(array, no_data_value)

    # drop empty geometries
    mask = ~geoseries.isnull()
    geoseries = geoseries[mask]
    values = values[mask] if values is not None else None

    # be strict about the bbox, it may lead to segfaults else
    x1, y1, x2, y2 = bbox
    if not ((x2 == x1 and y2 == y1) or (x1 < x2 and y1 < y2)):
        raise ValueError("Invalid bbox ({})".format(bbox))

    # if the request is a point, we find the intersecting polygon and
    # "rasterize" it
    if x2 == x1 and y2 == y1:
        mask = geoseries.intersects(Point(x1, y1))
        if not mask.any():
            pass
        elif values is not None:
            array[:] = values[mask].iloc[-1]  # take the last one
        else:
            array[:] = True
        return _finalize_rasterize_result(array, no_data_value)

    # create an output datasource in memory
    driver = ogr.GetDriverByName(str("Memory"))
    burn_attr = str("BURN_IT")
    sr = get_sr(projection)

    # prepare in-memory ogr layer
    ds_ogr = driver.CreateDataSource(str(""))
    layer = ds_ogr.CreateLayer(str(""), sr)
    layer_definition = layer.GetLayerDefn()
    if ogr_dtype is not None:
        field_definition = ogr.FieldDefn(burn_attr, ogr_dtype)
        layer.CreateField(field_definition)

    iterable = (
        zip(geoseries, values) if values is not None else zip(geoseries, repeat(True))
    )
    for geometry, value in iterable:
        feature = ogr.Feature(layer_definition)
        feature.SetGeometry(ogr.CreateGeometryFromWkb(geometry.wkb))
        if ogr_dtype is not None:
            feature[burn_attr] = value
        layer.CreateFeature(feature)

    geo_transform = GeoTransform.from_bbox(bbox, height, width)
    dataset_kwargs = {
        "no_data_value": no_data_value,
        "projection": sr.ExportToWkt(),
        "geo_transform": geo_transform,
    }

    # ATTRIBUTE=BURN_ATTR burns the BURN_ATTR value of each feature
    if dtype == np.uint8:  # this is our boolean dtype
        options = []
    else:
        options = [str("ATTRIBUTE=") + burn_attr]

    with Dataset(array, **dataset_kwargs) as dataset:
        gdal.RasterizeLayer(dataset, (1,), layer, options=options)

    return _finalize_rasterize_result(array, no_data_value)


def safe_abspath(url, start=None):
    """Executes safe_file_url but only returns the path and not the protocol.
    """
    url = safe_file_url(url)
    _, path = url.split("://")
    return path


def safe_file_url(url, start=None):
    """Formats an URL so that it meets the following safety conditions:

    - the URL starts with file:// (else: raises NotImplementedError)
    - the path is absolute (relative paths are taken relative to
      geomodeling.root)
    - if geomodeling.strict_paths: the path has to be contained inside
      `start` (else: raises IOError)

    For backwards compatibility, geomodeling.root can be overriden using the
    'start' argument.
    """
    try:
        protocol, path = url.split("://")
    except ValueError:
        protocol = "file"
        path = url
    else:
        if protocol != "file":
            raise NotImplementedError('Unknown protocol: "{}"'.format(protocol))
    if start is not None:
        warnings.warn(
            "Using the start argument in safe_file_url is deprecated. Use the "
            "'geomodeling.root' in the dask config",
            DeprecationWarning,
        )
    else:
        start = config.get("geomodeling.root")

    if not os.path.isabs(path):
        if start is None:
            raise IOError(
                "Relative path '{}' provided but start was not given.".format(path)
            )
        abspath = os.path.abspath(os.path.join(start, path))
    else:
        abspath = os.path.abspath(path)
    strict = config.get("geomodeling.strict-file-paths")
    if strict and not abspath.startswith(start):
        raise IOError("'{}' is not contained in '{}'".format(path, start))
    return "://".join([protocol, abspath])


PERCENTILE_REGEX = re.compile(r"^p([\d.]+)$")  # regex to match e.g. 'p50'


def parse_percentile_statistic(statistic):
    """Parses p<float> to a float, or None if it does not match."""
    # interpret percentile statistic
    percentile_match = PERCENTILE_REGEX.findall(statistic)
    if percentile_match:
        percentile = float(percentile_match[0])
        if not 0 <= percentile <= 100:
            raise ValueError("Percentiles must be in the range [0, 100]")
        return percentile


def dtype_for_statistic(dtype, statistic):
    """Return the result dtype of given statistic function"""
    # the dtype depends on the statistic
    if statistic in ("min", "max"):
        return dtype
    elif statistic == "sum":
        # same logic as Add block
        if np.issubdtype(dtype, np.integer) or dtype == np.bool:
            # use at least int32
            return np.result_type(dtype, np.int32)
        elif np.issubdtype(dtype, np.floating):
            # use at least float32
            return np.result_type(dtype, np.float32)
        else:
            return dtype
    elif statistic == "count":
        return np.int32
    else:
        # same logic as Divide block
        return np.result_type(np.float32, dtype)


def snap_start_stop(start, stop, time_first, time_delta, length):
    """This function interprets 'start' and 'stop' request parameters and
    returns the actual 'start' 'stop' with corresponding index range.

    There are 3 variants:
     - start and stop are None: the last frame is returned
     - only stop is None: the frame closest to start is returned
     - both start and stop are not None: the frames in the given interval are
       returned. The interval is closed on both sides.

    :param start: the requested 'start'
    :param stop: the requested 'stop'
    :param time_first: the timestamp of the first frame
    :param time_delta: the timedelta between consecutive frames
    :param length: the number of frames

    :type start: naive datetime
    :type stop: naive datetime
    :type time_first: naive datetime
    :type time_delta: timedelta or NoneType
    :type length: int
    """
    if length == 0:
        return (None,) * 4

    if length == 1:
        time_delta = None
        period = (time_first, time_first)
    elif length > 1 and time_delta is None:
        raise ValueError("Length > 1 requires a timedelta")
    else:
        period = (time_first, time_first + (length - 1) * time_delta)

    if start is None:
        # take the latest
        start = stop = period[-1]
        first_i = last_i = length - 1
    elif stop is None:
        # snap 'start' to the closest
        if start <= period[0]:
            start = stop = period[0]
            first_i = last_i = 0
        elif start >= period[1]:
            start = stop = period[1]
            first_i = last_i = length - 1
        elif length == 1:
            start = stop = period[0]
            first_i = last_i = 0
        else:
            # snap 'start' to the nearest frame
            first_i = last_i = int(round((start - period[0]) / time_delta))
            start = stop = period[0] + time_delta * first_i
    else:
        if start > period[1] or stop < period[0]:
            start = stop = first_i = last_i = None
        elif length == 1:
            start = stop = period[0]
            first_i = last_i = 0
        else:
            # ceil 'start'
            first_i = int(np.ceil((start - period[0]) / time_delta))
            first_i = max(first_i, 0)
            # floor 'stop'
            last_i = int(np.floor((stop - period[0]) / time_delta))
            last_i = min(last_i, length - 1)
            start = period[0] + time_delta * first_i
            stop = period[0] + time_delta * last_i
    return start, stop, first_i, last_i


def zoom_raster(data, no_data_value, height, width):
    """Zooms a data array to specified height and width

    Deals with no_data by setting these to 0, zooming, and putting back nodata.
    Edges around nodata will be biased towards 0."""
    if data.shape[1:] == (height, width):
        return data
    factor = 1, height / data.shape[1], width / data.shape[2]

    # first zoom the nodata mask
    src_mask = data == no_data_value
    dst_mask = ndimage.zoom(src_mask.astype(np.float), factor) > 0.5

    # set nodata to 0 and zoom the data
    data = data.copy()
    data[src_mask] = 0
    result = ndimage.zoom(data, factor)

    # set nodata to nodata again using the zoomed mask
    result[dst_mask] = no_data_value
    return result


def dt_to_ms(dt):
    """Converts a datetime to a POSIX timestamp in milliseconds"""
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=pytz.UTC)
    return int(dt.timestamp() * 1000)



"""
Module containing the base geometry block classes.
"""
import pandas as pd
from dask_geomodeling import Block

__all__ = ["GeometryBlock", "GetSeriesBlock", "SetSeriesBlock"]


class GeometryBlock(Block):
    """ The base block for geometries

    All geometry blocks must be derived from this base class and must implement
    the following attribute:

    - ``columns``: a set of column names to expect in the dataframe

    A geometry request contains the following fields:

    - mode: one of ``{"intersects", "centroid", "extent"}``
    - geometry: limit returned objects to objects that intersect with this
      shapely geometry object
    - projection: projection to return the geometries in as WKT string
    - limit: the maximum number of geometries
    - min_size: geometries with a bbox that is smaller than this on all sides
      are left out
    - start: start date as UTC datetime
    - stop: stop date as UTC datetime
    - filters: dict of `Django <https://www.djangoproject.com/>`_ ORM-like
      filters on properties (e.g. ``id=598``)

    The data response is a dictionary with the following fields:

    - (if mode was ``"intersects"`` or ``"centroid"``) ``"features"``:
      a ``GeoDataFrame`` of features with properties
    - (if mode was ``"extent"``) ``"extent"``: a tuple of 4 numbers
      ``(min_x, min_y, max_x, max_y)`` that represents the extent of the
      geometries that would be returned by an ``"intersects"`` request.
    - (for all modes) ``"projection"``: the EPSG or WKT representation of the
      projection.

    To be able to perform operations on properties, there is a helper type
    called ``SeriesBlock``. This is the block equivalent of a ``pandas.Series``.
    You can get a ``SeriesBlock`` from a ``GeometryBlock``, perform operations
    on it, and set it back into a ``GeometryBlock``.
    """

    def __getitem__(self, name):
        return GetSeriesBlock(self, name)

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError("Please use block.set to set a column.")

    def set(self, *args):
        # NB cannot use __setitem__ as block instances are immutable
        return SetSeriesBlock(self, *args)

    def to_file(self, *args, **kwargs):
        """Utility function to export data from this block to a file on disk.

        You need to specify the target file path as well as the extent geometry
        you want to save. Feature properties can be saved by providing a field
        mapping to the ``fields`` argument.

        To stay within memory constraints or to parallelize an operation, the
        ``tile_size`` argument can be provided.

        Args:
          url (str): The target file path. The extension determines the format.
            For supported formats, consult
            GeometryFileSink.supported_extensions.
          fields (dict): a mapping that relates column names to output file
            field names field names,
            ``{<output file field name>: <column name>, ...}``.
          tile_size (int): Optionally use this for large exports to stay within
            memory constraints. The export is split in tiles of given size
            (units are determined by the projection). Finally the tiles are
            merged.
          geometry (shapely Geometry): Limit exported objects to objects whose
            centroid intersects with this geometry.
          projection (str): The projection as a WKT string or EPSG code.
            Sets the projection of the geometry argument, the target
            projection of the data, and the tiling projection.
          start (datetime): start date as UTC datetime
          stop (datetime): stop date as UTC datetime
          **request: see GeometryBlock request specification

        Relevant settings can be adapted as follows:
          >>> from dask import config
          >>> config.set({"geomodeling.root": '/my/output/data/path'})
          >>> config.set({"geomodeling.geometry-limit": 10000})
          >>> config.set({"temporary_directory": '/my/alternative/tmp/dir'})
        """
        from dask_geomodeling.geometry.sinks import to_file

        return to_file(self, *args, **kwargs)


class SeriesBlock(Block):
    """ A block that represents one column from a GeometryBlock.

    Use this helper class to modify (or to use logic on) a specific feature
    property.

    Use :class:``dask_geomodeling.geometry.base.GetSeriesBlock`` to retrieve
    a SeriesBlock from a GeometryBlock and
    :class:``dask_geomodeling.geometry.base.SetSeriesBlock`` to add a
    SeriesBlock to a GeometryBlock.
    """

    def __add__(self, other):
        from . import Add

        return Add(self, other)

    def __mul__(self, other):
        from . import Multiply

        return Multiply(self, other)

    def __neg__(self):
        from . import Multiply

        return Multiply(self, -1)

    def __sub__(self, other):
        from . import Subtract

        return Subtract(self, other)

    def __floordiv__(self, other):
        from . import FloorDivide

        return FloorDivide(self, other)

    def __mod__(self, other):
        from . import Modulo

        return Modulo(self, other)

    def __truediv__(self, other):
        from . import Divide

        return Divide(self, other)

    def __pow__(self, other):
        from . import Power

        return Power(self, other)

    def __eq__(self, other):
        from . import Equal

        return Equal(self, other)

    def __ne__(self, other):
        from . import NotEqual

        return NotEqual(self, other)

    def __gt__(self, other):
        from . import Greater

        return Greater(self, other)

    def __ge__(self, other):
        from . import GreaterEqual

        return GreaterEqual(self, other)

    def __lt__(self, other):
        from . import Less

        return Less(self, other)

    def __le__(self, other):
        from . import LessEqual

        return LessEqual(self, other)

    def __invert__(self):
        from . import Invert

        return Invert(self)

    def __and__(self, other):
        from . import And

        return And(self, other)

    def __or__(self, other):
        from . import Or

        return Or(self, other)

    def __xor__(self, other):
        from . import Xor

        return Xor(self, other)


class GetSeriesBlock(SeriesBlock):
    """
    Obtain a single feature property column from a GeometryBlock.
    
    Provide a GeometryBlock with one or more columns. One of these columns can 
    be read from this source into a SeriesBlock. This SeriesBlock can be used
    to run for example classifications.

    Args:
      source (GeometryBlock): GeometryBlock with the column you want to load 
        into the SeriesBlock.
      name (str): Name of the column to load into the SeriesBlock.

    Returns: 
      SeriesBlock containing the single property column

    """

    def __init__(self, source, name):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(name, str):
            raise TypeError("'{}' object is not allowed".format(type(name)))
        if name not in source.columns:
            raise KeyError("Column '{}' is not available".format(name))
        super().__init__(source, name)

    @property
    def source(self):
        return self.args[0]

    @staticmethod
    def process(data, name):
        if "features" not in data or name not in data["features"].columns:
            return pd.Series([], dtype=float)
        return data["features"][name]


class SetSeriesBlock(GeometryBlock):
    """
    Add one or multiple property columns (SeriesBlocks) to a GeometryBlock.
    
    Provide the GeometryBlock that you want to add more properties to. Then
    provide the SeriesBlock(s) which you want to add to the GeometryBlock. The
    values of the SeriesBlock will be added to the features in the
    GeometryBlock automatically (if they are derived from the same geometries
    in previous operations, the features will have matching indexes so that
    each property is matched to the correct feature).

    The value which is set can also be a single value, in which case each
    feature will get the same value as a property.

    Args:
      source (GeometryBlock): The base GeometryBlock to which the SeriesBlock
        is added as a new column.
      column (str): The name of the new column (if it exists, it will be
        overwritten)
      value (SeriesBlock, number, str, bool): The SeriesBlock or constant value
        that has to be inserted in the destination column.
      *args: It is possible to repeat the ``"column"`` and ``"value"``
        arguments multiple times to insert more than one column.

    Example:
      Add two columns to an existing ``view`` like this:
      ``SetSeriesBlock(view, "column_1", series_1, "column_2", series_2)``.

    Returns:
      The source GeometryBlock with additional property columns

    """

    def __init__(self, source, column, value, *args):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        args = (column, value) + args
        if len(args) % 2 != 0:
            raise ValueError("The number of arguments should be even")
        for column in args[::2]:
            if not isinstance(column, str):
                raise TypeError("'{}' object is not allowed".format(type(column)))
        super().__init__(source, *args)

    @property
    def source(self):
        return self.args[0]

    @property
    def columns(self):
        return self.source.columns | set(self.args[1::2])

    @staticmethod
    def process(data, *col_val_pairs):
        if "features" not in data or len(data["features"]) == 0:
            return data
        features = data["features"].copy()
        for column, value in zip(col_val_pairs[::2], col_val_pairs[1::2]):
            features[column] = value

        return {"features": features, "projection": data["projection"]}


class BaseSingle(GeometryBlock):
    """Baseclass for all geometry blocks that adjust a geometry source"""

    def __init__(self, source, *args):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        super().__init__(source, *args)

    @property
    def source(self):
        return self.args[0]

    @property
    def columns(self):
        return self.source.columns  # by default, columns is not changed


class BaseSingleSeries(SeriesBlock):
    """Baseclass for all series blocks that adjust a single series source"""

    def __init__(self, source, *args):
        if not isinstance(source, SeriesBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        super().__init__(source, *args)

    @property
    def source(self):
        return self.args[0]



"""
Module containing operations that return series from geometry fields
"""

import pandas as pd

from dask_geomodeling.utils import shapely_transform

from .base import SeriesBlock, GeometryBlock


__all__ = ["Area"]


class Area(SeriesBlock):
    """
    Calculate the area of features in a GeometryBlock.

    Provide a GeometryBlock and a projection. Returns the area of each
    individual geometry in the input block, in that projection.
    
    Args:
      source (GeometryBlock): Source GeometryBlock which contains the features.
      projection (str): Projection in which to compute the area
        (i.e. ``"epsg:28992"``).
      
    Returns: 
      SeriesBlock with only the computed area
    """

    def __init__(self, source, projection):
        if not isinstance(source, GeometryBlock):
            raise TypeError("'{}' object is not allowed".format(type(source)))
        if not isinstance(projection, str):
            raise TypeError("Argument 'projection' must be a str.")
        super().__init__(source, projection)

    @property
    def source(self):
        return self.args[0]

    @property
    def projection(self):
        return self.args[1]

    @staticmethod
    def process(data, projection):
        if "features" not in data or len(data["features"]) == 0:
            return pd.Series([], dtype=float)

        return (
            data["features"]
            .geometry.apply(shapely_transform, args=(data["projection"], projection))
            .area
        )



from setuptools import setup
import os

version = '2.2.8'

long_description = "\n\n".join([open("README.rst").read(), open("CHANGES.rst").read()])

install_requires = (
    [
        "dask[delayed]>=0.18",
        "pandas>=0.19",
        "geopandas>=0.4",
        "pytz",
        "numpy>=1.12",
        "scipy>=0.19",
    ],
)

# emulate "--no-deps" on the readthedocs build (there is no way to specify this
# behaviour in the .readthedocs.yml)
if os.environ.get('READTHEDOCS') == 'True':
    install_requires = []


tests_require = ["pytest"]

setup(
    name="dask-geomodeling",
    version=version,
    description="On-the-fly operations on geographical maps.",
    long_description=long_description,
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    keywords=["dask"],
    author="Casper van der Wel",
    author_email="casper.vanderwel@nelen-schuurmans.nl",
    url="https://github.com/nens/dask-geomodeling",
    license="BSD 3-Clause License",
    packages=[
        "dask_geomodeling",
        "dask_geomodeling.core",
        "dask_geomodeling.geometry",
        "dask_geomodeling.raster",
    ],
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    python_requires='>=3.5',
    extras_require={"test": tests_require, "cityhash": ["cityhash"]},
    entry_points={"console_scripts": []},
)



import dask.config
import os

defaults = {
    "root": os.getcwd(),
    "strict-file-paths": False,
    "raster-limit": 12 * (1024 ** 2),  # ca. 100 MB of float64
    "geometry-limit": 10000,
}

dask.config.update_defaults({"geomodeling": defaults})
